

from __future__ import generators
from mapping import *


class QueryMatchWrapper(dict):
    """build a queryMatch mapping on demand, since its not actually needed
    during query traversal"""

    def __init__(self, dataMatch, compiler):
        dict.__init__(self)
        for k, v in dataMatch.items(): # INVERT THE DATA MATCH MAPPING
            self[v] = k
        for i in range(compiler.n): # ALSO SAVE MAPPINGS TO DATA EDGES
            gqi = compiler.gqi[i]
            self[gqi.fromNode, gqi.queryNode] = compiler.dataEdge[i]


class QueryMatchDescriptor(object):

    def __get__(self, obj, objtype):
        return QueryMatchWrapper(obj.dataMatch, obj)


class QueryMatcher(object):
    "map a query node or edge on demand"

    def __init__(self, compiler):
        self.compiler = compiler

    def __getitem__(self, k):
        for q, d in self.iteritems():
            if q == k:
                return d
        return KeyError

    def __iter__(self, k):
        for k, v in self.iteritems():
            yield k

    def iteritems(self):
        for dataNode, queryNode in self.compiler.dataMatch.items():
            yield queryNode, dataNode # RETURN NODE MAPPINGS
        for i in range(self.compiler.n): # ALSO SAVE MAPPINGS TO DATA EDGES
            gqi = self.compiler.gqi[i] # RETURN EDGE MAPPINGS
            yield (gqi.fromNode, gqi.queryNode), self.compiler.dataEdge[i]

    def items(self):
        return [x for x in self.iteritems()]

    def __repr__(self):
        return '{' + ','.join([repr(k) + ':' + repr(v)
                               for k, v in self.iteritems()]) + '}'


class GraphQueryCompiler(object):
    'compile a series of GraphQueryIterators into python code, run them'
    #queryMatch = QueryMatchDescriptor()
    _lang = "" # NO LANGUAGE STEM

    def __init__(self, name='graphquery', globalDict=None):
        self.name = name
        self.code = []
        self.unmark_code = []
        self.next_code = []
        self.end_code = []
        self.indent = []
        self.gqi = []
        self.queryLayerGraph = dictGraph()
        self.n = 0
        if globalDict is None:
            self._compiled = {}
        else:
            self._compiled = globalDict
        self.queryMatch = QueryMatcher(self)

    def __getitem__(self, key):
        'return appropropriate code for accessing nodes/edges in data or query'
        if key == 'n':
            return self.n
        elif key == 'name':
            return self.name
        elif key == 'dataGraph':
            queryEdge = self.gqi[self.n].queryGraph[self.gqi[self.n].fromNode]\
                    [self.gqi[self.n].queryNode]
            try: # CHECK IF QUERY EDGE USES A NON-DEFAULT DATA GRAPH
                dg = queryEdge['dataGraph']
                return 'self.gqi[%d].dataGraph' % self.n
            except (TypeError, KeyError):
                return 'dataGraph'
        elif key == 'filter':
            return 'self.gqi[%d].filter' % self.n
        elif key == 'toQueryNode':
            return 'self.gqi[%d].queryNode' % self.n
        elif key == 'fromQueryNode':
            return 'self.gqi[%d].fromNode' % self.n
        if key[:2] == 'to':
            layer = self.queryLayerGraph[self.gqi[self.n].queryNode]
        elif key[:4] == 'from':
            layer = self.queryLayerGraph[self.gqi[self.n].fromNode]
        if key[-8:] == 'DataNode': # GET 1ST LAYER WHERE THIS NODE ASSIGNED
            return 'dataNode%d' % layer.values()[0]
        if key[-8:] == 'DataEdge': # GET LAST LAYER, WHERE THIS EDGE CREATED
            return 'dataEdge[%d]' % layer.values()[-1]
        if key == 'level':
            return self.n
        try:
            return getattr(self.gqi[self.n], key)
        except AttributeError:
            raise KeyError('%s not a valid GraphQueryCompiler key' % key)

    def indent_code(self, codestr, current_indent):
        'calculate indentation levels added by code in codestr'
        codestr = codestr % self # PERFORM MACRO SUBSTITUTIONS
        lines = codestr.split('\n')
        lastline = lines[-1]
        if lastline == '' and len(lines) > 1: # IGNORE TERMINAL BLANK LINE
            lastline = lines[-2]
        # Determine final indentation level.
        nindent = len(lastline.split('\t')) - 1
        if len(lastline) > 0 and lastline[-1] == ':':
            nindent += 1
        s = '' # NOW FORMAT THE CODE WITH PROPER INDENTATION LEVEL
        for line in lines:
            s += current_indent * '\t' + line + '\n'
        return s, current_indent + nindent

    def __iadd__(self, gqi):
        'add a GraphQueryIterator to be compiled into this query'
        self.gqi.append(gqi)
        if gqi.queryNode not in self.queryLayerGraph: # NOT ALREADY BOUND?
            self.queryLayerGraph += gqi.queryNode
            codestr = getattr(gqi, self._lang + '_generator_code')
            markstr = getattr(gqi, self._lang + '_index_code')
            unmarkstr = getattr(gqi, self._lang + '_unmark_code')
            try:
                endcode = getattr(gqi, self._lang + '_end_code')
            except AttributeError:
                endcode = ''
            try:
                nextcode = getattr(gqi, self._lang + '_next_code')
            except AttributeError:
                nextcode = ''
            self.lastGenerator = self.n
        else:
            codestr = getattr(gqi, self._lang + '_closure_code')
            markstr = None
            unmarkstr = getattr(gqi, self._lang + '_unmark_closure_code')
            try:
                endcode = getattr(gqi, self._lang + '_end_closure_code')
            except AttributeError:
                endcode = ''
            try:
                nextcode = getattr(gqi, self._lang + '_next_closure_code')
            except AttributeError:
                nextcode = ''
        #BIND QUERY EDGE TO THIS LAYER
        self.queryLayerGraph[gqi.queryNode][gqi.fromNode] = self.n
        try: # GET INDENTATION LEVEL FROM PREVIOUS LAYER
            current_indent = self.indent[-1]
        except IndexError:
            current_indent = 1 # TOPLEVEL: MUST INDENT INSIDE def
        self.end_code.append(self.indent_code(endcode, current_indent)[0])
        s, current_indent = self.indent_code(codestr, current_indent)
        self.next_code.append(self.indent_code(nextcode, current_indent)[0])
        if hasattr(gqi, 'filter'):
            s2, current_indent = self.indent_code(getattr(gqi, self._lang + \
                                                          '_filter_code'),
                                                  current_indent)
            s += s2
        if hasattr(gqi, 'filtercode'):
            s2, current_indent = self.indent_code(gqi.filtercode,
                                                  current_indent)
            s += s2
        if markstr is not None:
            s2, current_indent = self.indent_code(markstr, current_indent)
            s += s2
        if unmarkstr is not None:
            s2, tail_indent = self.indent_code(unmarkstr, current_indent)
            self.unmark_code.append(s2)
        else:
            self.unmark_code.append('') # NO UNMARK CODE, SO JUST APPEND BLANK
        self.code.append(s)
        self.indent.append(current_indent)
        self.n += 1
        return self # iadd MUST RETURN self!!
    _def_code = """
def %(name)s(self, dataGraph, dataMatch=None, queryMatch=None):
\tif dataMatch is None: dataMatch={}
\tself.dataMatch = dataMatch
\tdataEdge = %(n)d * [None]
\tself.dataEdge = dataEdge
"""
    _yield_code = 'yield self.queryMatch\n'
    _end_code = ''

    def __str__(self):
        'generate code for this query, as a string function definition'
        s = self._def_code % self
        for layer in self.code: # GENERATE ALL THE TRAVERSAL CODE
            s += layer
        # yield the result
        s2 = self.indent_code(self._yield_code, self.indent[-1])[0]
        s += s2
        i = len(self.unmark_code) - 1
        while i >= 0: # GENERATE THE UNMARKING CODE...
            s += self.unmark_code[i]
            s += self.next_code[i]
            s += self.end_code[i]
            i -= 1
        s += self._end_code % self
        return s

    def run(self, dataGraph, *args, **kwargs):
        'run the query, pre-compiling it if necessary'
        try: # JUST TRY RUNNING OUR FUNCTION: IT RETURNS AN ITERATOR
            return self._compiled[self.name](self, dataGraph, *args, **kwargs)
        except KeyError:
            self.compile()
            # Run it.
            return self._compiled[self.name](self, dataGraph, *args, **kwargs)

    def compile(self):
        'compile using Python exec statement'
        exec str(self) in self._compiled # COMPILE OUR FUNCTION


def find_distutils_lib(path='build'):
    'locate the build/lib path where distutils builds modules'
    import os
    dirs = os.listdir('build')
    for d in dirs:
        if d[:4] == 'lib.':
            return path + '/' + d
    raise OSError((1, 'Unable to locate where distutils built your module!'))


class GraphQueryPyrex(GraphQueryCompiler):
    'compile a series of GraphQueryIterators into pyrex code, run them'
    #queryMatch = QueryMatchDescriptor()
    _lang = "_pyrex" # NO LANGUAGE STEM

    def __getitem__(self, key):
        'return appropropriate code for accessing nodes/edges in data or query'
        if key == 'n':
            return self.n
        elif key == 'name':
            return self.name
        elif key == 'dataGraph':
            try: # CHECK IF QUERY EDGE USES A NON-DEFAULT DATA GRAPH
                queryEdge = self.gqi[self.n].queryGraph[self.gqi[self.n].\
                                          fromNode][self.gqi[self.n].queryNode]
                dg = queryEdge['dataGraph']
                return 'self.gqi[%d].dataGraph' % self.n
            except (TypeError, KeyError):
                return 'dataGraph'
        elif key == 'filter':
            return 'self.gqi[%d].filter' % self.n
        elif key == 'toQueryNode':
            return 'self.gqi[%d].queryNode' % self.n
        elif key == 'fromQueryNode':
            return 'self.gqi[%d].fromNode' % self.n
        if key[:2] == 'to':
            layer = self.queryLayerGraph[self.gqi[self.n].queryNode]
        elif key[:4] == 'from':
            layer = self.queryLayerGraph[self.gqi[self.n].fromNode]
        if key[-8:] == 'DataNode': # GET 1ST LAYER WHERE THIS NODE ASSIGNED
            return 'dataNode%d' % layer.values()[0]
        if key[-8:] == 'DataEdge': # GET LAST LAYER, WHERE THIS EDGE CREATED
            return 'dataEdge%d' % layer.values()[-1]
        if key[-8:] == 'DataDict': # GET 1ST LAYER WHERE THIS NODE ASSIGNED
            return 'cDict%d' % layer.values()[0]
        if key[-7:] == 'DataPtr': # GET 1ST LAYER WHERE THIS NODE ASSIGNED
            return 'pDictEntry%d' % layer.values()[0]
        if key[-11:] == 'DataPtrCont': # GET 1ST LAYER WHERE THIS NODE ASSIGNED
            return 'pGraphEntry%d' % layer.values()[0]
        if key[-11:] == 'DataCounter': # GET 1ST LAYER WHERE THIS NODE ASSIGNED
            return 'i%d' % layer.values()[0]
        if key == 'toDataNodeUnmatched':
            l = ['dataNode%d!=dataNode%d'
                 % (self.queryLayerGraph[self.gqi[i].queryNode].values()[0],
                  self.queryLayerGraph[self.gqi[self.n].queryNode].values()[0])
               for i in range(self.n)]
            if len(l) > 0:
                return ' and '.join(l)
            else:
                return 'True'
        if key == 'dataNodeDefs':
            return ','.join(['dataNode%d' % i for i in range(self.n)])
        if key == 'dataEdgeDefs':
            return ','.join(['dataEdge%d' % i for i in range(self.n)])
        if key == 'dataDictDefs':
            return ','.join(['*cDict%d' % i for i in range(self.n)])
        if key == 'dataPtrDefs':
            return ','.join(['*pDictEntry%d' % i for i in range(self.n)])
        if key == 'dataPtrContDefs':
            return ','.join(['*pGraphEntry%d' % i for i in range(self.n)])
        if key == 'dataCounterDefs':
            return ','.join(['i%d' % i for i in range(self.n)])
        if key == 'dataCounterArgs':
            return ','.join(['int i%d' % i for i in range(self.n)])
        if key == 'itaVector':
            return ','.join(['ita.vector[%d]' % i for i in range(self.n)])
        if key == 'itaTuple':
            return ',\\\n'.join(['p_ita[%d]' % i for i in range(2 * self.n)])
        if key == 'resultTuple':
            return ',\\\n'.join(['dataNode%d,dataEdge%d'
                 % (self.queryLayerGraph[self.gqi[i].queryNode].values()[0], i)
                                 for i in range(self.n)])
        if key == 'resultTuples':
            return ','.join(['(dataNode%d,dataEdge%d)'
                 % (self.queryLayerGraph[self.gqi[i].queryNode].values()[0], i)
                             for i in range(self.n)])
        if key == 'level' or key == 'nEdges':
            return self.n
        if key == 'lastGenerator':
            return self.lastGenerator
        try:
            return getattr(self.gqi[self.n], key)
        except AttributeError:
            raise KeyError('%s not a valid GraphQueryPyrex key' % key)

    _def_code = """
cimport cdict
cdef c_%(name)s(cdict.CGraphDict cgd, cdict.IntTupleArray ita,
                    %(dataCounterArgs)s):
\tcdef cdict.CGraph *dataGraph
\tcdef cdict.CDict %(dataDictDefs)s
\tcdef cdict.CDictEntry %(dataPtrDefs)s
\tcdef cdict.CDictEntry *pd_temp
\tcdef cdict.CGraphEntry %(dataPtrContDefs)s
\tcdef cdict.CGraphEntry *p_temp
\t#cdef int %(dataCounterDefs)s
\tcdef int %(dataNodeDefs)s
\tcdef int %(dataEdgeDefs)s
\tcdef int *p_ita
\tdataGraph = cgd.d
\tp_ita = ita.data
"""
    _yield_code = """
%(itaTuple)s = %(resultTuple)s
p_ita = p_ita + 2 * %(nEdges)d
ita.n = ita.n + 1
if ita.n >= ita.n_alloc:
\tita.set_vector((%(dataCounterDefs)s), 1)
\treturn
"""
    #results.append((%(resultTuples)s))\n
    _end_code = """
\tita.isDone = 1 # COMPLETED THIS QUERY

from pygr import cdict
def %(name)s(self, g, int maxhit=1000, cdict.IntTupleArray ita=None, qml=None):
\tif not isinstance(g, cdict.CGraphDict):
\t\tg = cdict.CGraphDict(g, cdict.KeyIndex())
\tif ita is None:
\t\tita = cdict.IntTupleArray(maxhit, %(nEdges)d, 2, %(lastGenerator)d)
\tita.n = 0
\tc_%(name)s(g, ita, %(itaVector)s) # RUN THE QUERY
\tif qml is not None:
\t\tqml.matches = ita
\t\treturn qml
\telse:
\t\treturn cdict.QueryMatchList(self, ita, g, %(name)s)
"""

    def compile(self):
        'compile using Pyrex, Distutils, and finally import!'
        import os
        try:
            # We need access to Pygr source code to access cgraph functions
            # in this module.
            pygrpath = os.environ['PYGRPATH']
        except KeyError:
            raise OSError((1,
                           """pyrex compilation requires access to pygr source.
                           Please set the environment variable PYGRPATH to \
                           the top of the pygr source package."""))
        if not os.access(pygrpath + '/pygr/cgraph.c', os.R_OK):
            raise OSError((1, """Unable to access %s/pygr/cgraph.c.
            Is PYGRPATH set to the top of the pygr source package?"""
                           % pygrpath))
        exit_status = os.system('cp %s/pygr/cgraph.c %s/pygr/cgraph.h \
                                %s/pygr/cdict.pxd .'
                                % (pygrpath, pygrpath, pygrpath))
        if exit_status != 0:  # RUN THE PYREX COMPILER TO PRODUCE C
            raise OSError((exit_status,
                           'unable to copy source code to this directory.'))
        # Construct a unique name for the module.
        modulename = self.name + str(id(self))
        myfile = file(modulename + '.pyx', 'w') # GENERATE PYREX CODE
        myfile.write(str(self)) # WRITE CODE
        myfile.close()
        exit_status = os.system('pyrexc %s.pyx' % (modulename))
        if exit_status != 0:  # RUN THE PYREX COMPILER TO PRODUCE C
            raise OSError((exit_status, 'pyrex compilation failed. Is \
pyrexc missing or not in your PATH?'))
        # Build the module using distutils.
        from distutils.core import setup, Extension
        module1 = Extension(modulename, sources=['cgraph.c',
                                                 modulename + '.c'])
        setup(name=modulename,
              description='autogenerated by pygr.graphquery',
              ext_modules=[module1], script_args=['build'])
        # Find out where distutils put our built module.
        modulepath = find_distutils_lib()
        # Work around a nasty problem with Pyrex cimport - there is no way to
        # tell it the module is in a subdirectory! Here, 'from pygr cimport
        # cdict' or 'cimport pygr.cdict' fail; one MUST say 'cimport cdict'.
        import sys
        import os.path
        from pygr import cdict
        # Add the module's location to our path.
        sys.path += [os.path.dirname(cdict.__file__)]
        import imp  # FINALLY, TRY TO IMPORT THE NEW MODULE
        modulefile, path, desc = imp.find_module(modulename, [modulepath])
        # Load and bind the module.
        self._module = imp.load_module(modulename, modulefile, path, desc)
        # Bind our query function.
        self._compiled[self.name] = getattr(self._module, self.name)
        modulefile.close()


class GraphQueryIterator(object):
    """iterator for a single node in graph query.  Subclasses provide different
       flavors of generator methods: graph w/ edges; container; attr;
       function etc."""

    def __init__(self, fromNode, queryNode, dataGraph, queryGraph,
                 dataMatch, queryMatch, attrDict={}):
        self.fromNode = fromNode
        self.queryNode = queryNode
        self.dataGraph = dataGraph
        self.queryGraph = queryGraph
        self.dataMatch = dataMatch
        self.queryMatch = queryMatch
        self.dataNode = None
        for attr, val in attrDict.items():
            # Save our edge information as attributes of this object.
            setattr(self, attr, val)
##         try:
##             self.nq = len(self.queryGraph[self.queryNode])
##         except KeyError:
##             self.nq = 0

    def restart(self):
        "reset the iterator to its beginning"
        self.mustMark = True
        if self.fromNode != None:
            self.dataNode = self.queryMatch[self.fromNode]
        if self.queryNode in self.queryMatch: # ALREADY ASSIGNED TO A DATA NODE
            self.mustMark = False
            if self.fromNode is None: # NO PATH TO HERE, SO JUST ECHO SINGLETON
                self.iterator = self.echo()
            else: # CHECK FOR PATH FROM fromNode TO THIS DATA NODE
                self.iterator = self.closure()
        else:
            self.iterator = self.generate()

    def echo(self):
        "Just return what our node is ALREADY matched to"
        yield self.queryMatch[self.queryNode], None

    def closure(self):
        "This node is already matched. Make sure a path to it (closure) exists"
        targetNode = self.queryMatch[self.queryNode]
        try: # GENERATE IF container HAS EDGE TO targetNode
            container = self.dataGraph[self.dataNode]
            yield targetNode, container[targetNode]
        except KeyError:
            pass

    def generate(self):
        "generate all neighbors of data node matched to fromNode"
        try:
            it = self.dataGraph[self.dataNode]
        except KeyError:
            pass
        else:
            for i, e in it.items():
                yield i, e

    _generator_code = """
try: # GENERATOR
\tit%(level)d = %(dataGraph)s[%(fromDataNode)s]
except KeyError:
\tcontinue
for %(toDataNode)s, %(toDataEdge)s in it%(level)d.items():"""

    _index_code = """
if %(toDataNode)s in dataMatch:
\tcontinue
else:
\tdataMatch[%(toDataNode)s] = %(toQueryNode)s
\t#queryMatch[%(toQueryNode)s] = %(toDataNode)s
\t#queryMatch[%(fromQueryNode)s, %(toQueryNode)s] = %(toDataEdge)s
# THIS LINE PREVENTS COMPILER FROM PUSHING EXTRA INDENTATION LAYER"""

    _filter_code = """
if self.gqi[%(level)d].filter(toNode=%(toDataNode)s, fromNode=%(fromDataNode)s, \
                              edge=%(toDataEdge)s, queryMatch=self.queryMatch, \
                              gqi=self.gqi[%(level)d]):"""

    _unmark_code = """
del dataMatch[%(toDataNode)s]
#del queryMatch[%(toQueryNode)s]
#del queryMatch[%(fromQueryNode)s, %(toQueryNode)s]"""

    _closure_code = """
try: # CLOSURE
\t%(toDataEdge)s = %(dataGraph)s[%(fromDataNode)s][%(toDataNode)s]
except KeyError:
\tpass
else:
\t#queryMatch[%(fromQueryNode)s, %(toQueryNode)s] = %(toDataEdge)s"""

    _unmark_closure_code = """
#del queryMatch[%(fromQueryNode)s, %(toQueryNode)s]"""

    # PYREX CODE
    _pyrex_generator_code = """
p_temp = cdict.cgraph_getitem(%(dataGraph)s, %(fromDataNode)s)
if p_temp != NULL:
\t%(fromDataDict)s = p_temp[0].v
\t%(toDataPtr)s = %(fromDataDict)s[0].dict
\twhile %(toDataCounter)s < %(fromDataDict)s[0].n:
\t\t%(toDataNode)s = %(toDataPtr)s[%(toDataCounter)s].k
\t\t%(toDataEdge)s = %(toDataPtr)s[%(toDataCounter)s].v
"""
    #for %(toDataCounter)s from 0 <= %(toDataCounter)s < %(fromDataDict)s[0].n:

    _pyrex_index_code = 'if %(toDataNodeUnmatched)s:'

    _pyrex_unmark_code = '# COMPILER NEEDS AT LEAST ONE LINE, \
                          EVEN THOUGH NOTHING TO DO HERE'

    _pyrex_next_code = '%(toDataCounter)s = %(toDataCounter)s + 1'

    _pyrex_end_code = '%(toDataCounter)s = 0'

    _pyrex_closure_code = """
p_temp = cdict.cgraph_getitem(%(dataGraph)s, %(fromDataNode)s)
if p_temp != NULL:
\t%(fromDataDict)s = p_temp[0].v
\tpd_temp = cdict.cdict_getitem(%(fromDataDict)s, %(toDataNode)s)
\tif pd_temp != NULL:
\t\t%(toDataEdge)s = pd_temp[0].v
"""

    _pyrex_unmark_closure_code = '# COMPILER NEEDS AT LEAST ONE LINE, \
                                  EVEN THOUGH NOTHING TO DO HERE'

    def unmark(self):
        "erase node and edge assignment associated with the iterator"
        if self.mustMark and self.queryNode in self.queryMatch:
            i = self.queryMatch[self.queryNode] # ERASE OLD NODE ASSIGNMENT
            del self.dataMatch[i]
            del self.queryMatch[self.queryNode]
        try:
            # Erase old edge.
            del self.queryMatch[(self.fromNode, self.queryNode)]
        except KeyError:
            pass

    def next(self):
        "returns the next node from iterator that passes all tests"
        self.unmark()
        for i, e in self.iterator: # RETURN THE FIRST ACCEPTABLE ITEM
##             try:
##                 # Check the number of outgoing edges. NOTE: This check
##                 # will NOT work if multiple graphs are queried!
##                 nd = len(self.dataGraph[i])
##             except KeyError:
##                 nd = 0
##             if nd >= self.nq and
            if self.mustMark and i in self.dataMatch:
                continue # THIS NODE ALREADY ASSIGNED. CAN'T REUSE IT!
            if (not hasattr(self, 'filter') # APPLY EDGE / NODE TESTS HERE
               or self.filter(toNode=i, fromNode=self.dataNode, edge=e,
                               queryMatch=self.queryMatch, gqi=self)):
                if self.mustMark:
                    # Save this node assignment.
                    self.dataMatch[i] = self.queryNode
                    self.queryMatch[self.queryNode] = i
                if e is not None: # SAVE EDGE INFO, IF ANY
                    self.queryMatch[(self.fromNode, self.queryNode)] = e
                return i  # THIS ITEM PASSES ALL TESTS.  RETURN IT
        return None # NO MORE ITEMS FROM THE ITERATOR


class ContainerGQI(GraphQueryIterator):
    "Iterate over all nodes in self.dataGraph"

    def generate(self):
        for i in self.dataGraph:
            yield i, None

    _generator_code = """
%(toDataEdge)s = None # CONTAINER
for %(toDataNode)s in dataGraph:"""

    _pyrex_generator_code="""
%(toDataPtrCont)s = %(dataGraph)s[0].dict
for %(toDataCounter)s from 0 <= %(toDataCounter)s < %(dataGraph)s[0].n:
\t%(toDataNode)s = %(toDataPtrCont)s[%(toDataCounter)s].k
\t%(toDataEdge)s = -1 # NO EDGE INFO
"""


class AttributeGQI(GraphQueryIterator):
    "Iterate over all nodes in attribute called self.attr of self.dataNode"

    def generate(self):
        for i, e in getattr(self.dataNode, self.attr).items():
            yield i, e

    _generator_code = """
for %(toDataNode)s, %(toDataEdge)s in getattr(%(fromDataNode)s, \
                                              '%(attr)s').items():"""


class AttrContainerGQI(GraphQueryIterator):
    """Iterate over all nodes in attribute called self.attrN of self.dataNode
    (no edge info)"""

    def generate(self):
        for i in getattr(self.dataNode, self.attrN):
            yield i, None

    _generator_code = """
%(toDataEdge)s = None
for %(toDataNode)s in getattr(%(fromDataNode)s, '%(attrN)s'):"""


class CallableGQI(GraphQueryIterator):
    "Call the specified function self.f as iterator"

    def generate(self):
        for i, e in self.f(self.dataNode, self.dataGraph, self):
            yield i, e

    _generator_code = """
for %(toDataNode)s, %(toDataEdge)s in self.gqi[%(level)d].f(%(fromDataNode)s, \
                                            dataGraph, self.gqi[%(level)d]):"""


class CallableContainerGQI(GraphQueryIterator):
    "Call the specified function self.fN as iterator (no edge info)"

    def generate(self):
        for i in self.fN(self.dataNode, self.dataGraph, self):
            yield i, None

    _generator_code = """
%(toDataEdge)s = None
for %(toDataNode)s in self.gqi[%(level)d].fN(%(fromDataNode)s, dataGraph, \
                                             self.gqi[%(level)d]):"""


class SubqueryGQI(GraphQueryIterator):
    """base class for running subqueries; produces a union of all subquery
    solutions. self.subqueries must be list of graph objects, each
    representing a subquery"""

    def __init__(self, fromNode, queryNode, dataGraph, queryGraph,
                 dataMatch, queryMatch, attrDict={}):
        GraphQueryIterator.__init__(self, fromNode, queryNode, dataGraph,
                                    queryGraph, dataMatch, queryMatch,
                                    attrDict)
        self.graphQueries = []
        for qg in self.subqueries: # INITIALIZE OUR SUBQUERIES
            self.graphQueries.append(self.gqClass(self.dataGraph, qg,
                                                  dataMatch, queryMatch))

    def closure(self):
        "Generate union of all solutions returned by all subqueries"
        for gq in self.graphQueries:
            for d in gq: # LAUNCHES THE GRAPH QUERY, GETS ALL ITS SOLUTIONS
                yield self.queryMatch[self.queryNode], None
            # Remove its query-data mapping before going to the next subquery.
            gq.cleanup()


def newGQI(self, oclass, fromNode, toNode, dataGraph, queryGraph,
           dataMatch, queryMatch, gqiDict):
    """figure out a default GQI class to use, based on an attribute dictionary,
       then return a new object of that class initialized with the input data
       """
    if fromNode is not None and toNode is not None and \
           queryGraph[fromNode][toNode] is not None:
        kwargs = queryGraph[fromNode][toNode]
        for attr in kwargs:
            try:
                # Use attribute name to determine default class.
                oclass = gqiDict[attr]
            except KeyError:
                pass
    else:
        kwargs = {}
    try:
        oclass = kwargs['__class__'] # LET USER SET CLASS TO USE
    except KeyError:
        pass
    return oclass(fromNode, toNode, dataGraph, queryGraph, dataMatch,
                  queryMatch, kwargs)


class GraphQuery(object):
    "represents a single query or subquery"
    # DEFAULT MAPPING OF ATTRIBUTE NAMES TO GQI CLASSES TO USE WITH THEM
    gqiDict = {'attr': AttributeGQI,
               'attrN': AttrContainerGQI,
               'f': CallableGQI,
               'fN': CallableContainerGQI,
               'subqueries': SubqueryGQI}
    newGQI = newGQI # USE THIS METHOD TO CHOOSE GQI CLASS FOR EACH ITERATOR

    def __init__(self, dataGraph, queryGraph, dataMatch=None, queryMatch=None):
        """Enumerate nodes in queryGraph in BFS order,
        constructing iterator stack"""
        self.dataGraph = dataGraph
        self.queryGraph = queryGraph
        if dataMatch is None:
            dataMatch = {}
        if queryMatch is None:
            queryMatch = {}
        self.dataMatch = dataMatch
        self.queryMatch = queryMatch
        # First we need to find start nodes: process them first and mark them
        # as generate all.
        isFollower = {}
        for node in queryGraph:
            for node2 in queryGraph[node]:
                # node2 has an incoming edge so it cannot be a start node.
                isFollower[node2] = True
        q = []
        self.q = q
        n = 0
        for node in queryGraph: # PLACE START NODES AT HEAD OF QUEUE
            if node not in isFollower:
                q.append(self.newGQI(ContainerGQI, None, node, dataGraph,
                                     queryGraph, dataMatch, queryMatch,
                                     self.gqiDict))
                n += 1
        if n == 0:
            # No start nodes; just add the first query node to the queue.
            for node in queryGraph:
                q.append(self.newGQI(ContainerGQI, None, node, dataGraph,
                                     queryGraph, dataMatch, queryMatch,
                                     self.gqiDict))
                n += 1
                break # Only add the first node.
        if n == 0:
            raise ValueError('query graph is empty!')

        visited = {}
        i = 0
        while i < n:
            # Add node to the queue even if it's already been visited
            # - but don't add its neighbours.
            if q[i].queryNode not in visited: # ADD NEIGHBORS TO THE QUEUE
                visited[q[i].queryNode] = True # MARK AS VISITED
                for node in queryGraph[q[i].queryNode]: # GET ALL ITS NEIGHBORS
                    #print 'QUEUE:', n, node
                    q.append(self.newGQI(GraphQueryIterator, q[i].queryNode,
                                         node, dataGraph, queryGraph,
                                         dataMatch, queryMatch,
                                         self.gqiDict))
                    n += 1
            i += 1

    def __iter__(self):
        "generates all subgraphs of dataGraph matching queryGraph"
        i = 0
        n = len(self.q)
        self.q[0].restart() # PRELOAD ITERATOR FOR 1ST NODE
        while i >= 0:
            dataNode = self.q[i].next()
            if dataNode is not None:
                #print i,qu[i].queryNode,dataNode
                if i + 1 < n: # MORE LEVELS TO QUERY?
                    i += 1 # ADVANCE TO NEXT QUERY LEVEL
                    self.q[i].restart()
                else:  # GRAPH MATCH IS COMPLETE!
                    yield self.queryMatch # RETURN COMPLETE MATCH

            else: # NO MORE ACCEPTABLE NODES AT THIS LEVEL, SO BACKTRACK
                i -= 1

    def cleanup(self):
        "erase any query:data node matching associated with this subquery"
        for q in self.q:
            q.unmark()

    def compile(self, globals=None, compilerClass=GraphQueryCompiler,
                **kwargs):
        """return a compiled version of this query, using globals namespace
        if specified"""
        compiler = compilerClass(globalDict=globals, **kwargs)
        for gqi in self.q:
            compiler += gqi
        return compiler


SubqueryGQI.gqClass = GraphQuery # CLASS FOR CONSTRUCTING SUBQUERIES



import glob
import os
import tempfile
import classutil
import logger
from sequtil import *
from parse_blast import BlastHitParser
from seqdb import write_fasta, read_fasta
from nlmsa_utils import CoordsGroupStart, CoordsGroupEnd, CoordsToIntervals,\
     EmptySlice
from annotation import AnnotationDB, TranslationAnnot, TranslationAnnotSlice
import cnestedlist
import translationDB
import UserDict

# NCBI HAS THE NASTY HABIT OF TREATING THE IDENTIFIER AS A BLOB INTO
# WHICH THEY STUFF FIELD AFTER FIELD... E.G. gi|1234567|foobarU|NT_1234567|...
# THIS JUST YANKS OUT THE SECOND ARGUMENT SEPARATED BY |
NCBI_ID_PARSER=lambda id: id.split('|')[1]


def blast_program(query_type, db_type):
    progs = {DNA_SEQTYPE: {DNA_SEQTYPE: 'blastn', PROTEIN_SEQTYPE: 'blastx'},
            PROTEIN_SEQTYPE: {DNA_SEQTYPE: 'tblastn',
                              PROTEIN_SEQTYPE: 'blastp'}}
    if query_type == RNA_SEQTYPE:
        query_type = DNA_SEQTYPE
    if db_type == RNA_SEQTYPE:
        db_type = DNA_SEQTYPE
    return progs[query_type][db_type]


def read_blast_alignment(ofile, srcDB, destDB, al=None, pipeline=None,
                         translateSrc=False, translateDest=False):
    """Apply sequence of transforms to read input from 'ofile'.

    srcDB: database for finding query sequences from the blast input;

    destDB: database for finding subject sequences from the blast input;

    al, if not None, must be a writeable alignment object in which to
    store the alignment intervals;

    translateSrc=True forces creation of a TranslationDB representing
    the possible 6-frames of srcDB (for blastx, tblastx);

    translateDest=True forces creation of a TranslationDB representing
    the possible 6-frames of destDB (for tblastn, tblastx).

    If pipeline is not None, it must be a list of filter functions each
    taking a single argument and returning an iterator or iterable result
    object.
    """
    p = BlastHitParser()
    d = dict(id='src_id', start='src_start', stop='src_end', ori='src_ori',
             idDest='dest_id', startDest='dest_start',
             stopDest='dest_end', oriDest='dest_ori')
    if translateSrc:
        srcDB = translationDB.get_translation_db(srcDB)
    if translateDest:
        destDB = translationDB.get_translation_db(destDB)
    cti = CoordsToIntervals(srcDB, destDB, d)
    alignedIvals = cti(p.parse_file(ofile))
    if pipeline is None:
        result = save_interval_alignment(alignedIvals, al)
    else: # apply all the filters in our pipeline one by one
        result = alignedIvals
        for f in pipeline:
            result = f(result)
    return result


def save_interval_alignment(alignedIvals, al=None):
    """Save alignedIvals to al, or a new in-memory NLMSA"""
    needToBuild = False
    if al is None:
        al = cnestedlist.NLMSA('blasthits', 'memory', pairwiseMode=True,
                               bidirectional=False)
        needToBuild = True
    al.add_aligned_intervals(alignedIvals)
    if needToBuild:
        al.build()
    return al


def start_blast(cmd, seq, seqString=None, seqDict=None, **kwargs):
    """Run blast and return results."""
    p = classutil.FilePopen(cmd, stdin=classutil.PIPE, stdout=classutil.PIPE,
                            **kwargs)
    if seqString is None:
        seqString = seq
    if seqDict is not None: # write all seqs to nonblocking ifile
        for seqID, seq in seqDict.iteritems():
            write_fasta(p.stdin, seq)
        seqID = None
    else: # just write one query sequence
        seqID = write_fasta(p.stdin, seqString)
    if p.wait(): # blast returned error code
        raise OSError('command %s failed' % ' '.join(cmd))
    return seqID, p


def process_blast(cmd, seq, blastDB, al=None, seqString=None, queryDB=None,
                  popenArgs={}, **kwargs):
    """Run blast and return an alignment."""
    seqID, p = start_blast(cmd, seq, seqString, seqDict=queryDB, **popenArgs)
    try:
        if not queryDB: # need a query db for translation / parsing results
            try:
                queryDB = seq.db # use this sequence's database
            except AttributeError:
                queryDB = {seqID: seq} # construct a trivial "database"

        al = read_blast_alignment(p.stdout, queryDB, blastDB, al, **kwargs)
    finally:
        p.close() # close our PIPE files
    return al


def repeat_mask(seq, progname='RepeatMasker', opts=()):
    'Run RepeatMasker on a sequence, return lowercase-masked string'
    ## fd, temppath = tempfile.mkstemp()
    ## ofile = os.fdopen(fd, 'w') # text file
    p = classutil.FilePopen([progname] + list(opts), stdin=classutil.PIPE,
                            stdinFlag=None)
    write_fasta(p.stdin, seq, reformatter=lambda x: x.upper()) # save uppercase
    try:
        if p.wait():
            raise OSError('command %s failed' % ' '.join(p.args[0]))
        ifile = file(p._stdin_path + '.masked', 'rU') # text file
        try:
            for id, title, seq_masked in read_fasta(ifile):
                break # JUST READ ONE SEQUENCE
        finally:
            ifile.close()
    finally: # clean up our temp files no matter what happened
        p.close() # close temp stdin file
        for fpath in glob.glob(p._stdin_path + '.*'):
            try:
                os.remove(fpath)
            except OSError:
                pass
    return seq_masked # ONLY THE REPEATS ARE IN LOWERCASE NOW


def warn_if_whitespace(filepath):
    l = filepath.split() # check filepath for whitespace
    if len(l) > 1 or len(l[0]) < len(filepath): # contains whitespace
        logger.warn("""
Your sequence filepath contains whitespace characters:
%s
The NCBI formatdb (and blastall) programs cannot handle file paths
containing whitespace! This is a known NCBI formatdb / blastall bug.
Please use a path containing no whitespace characters!""" % filepath)
        return True # signal caller that a warning was issued


class BlastMapping(object):
    'Graph interface for mapping a sequence to homologs in a seq db via BLAST'

    def __init__(self, seqDB, filepath=None, blastReady=False,
                 blastIndexPath=None, blastIndexDirs=None, verbose=True,
                 showFormatdbMessages=True, **kwargs):
        '''seqDB: sequence database object to search for homologs
        filepath: location of FASTA format file for the sequence database
        blastReady: if True, ensure that BLAST index file ready to use
        blastIndexPath: location of the BLAST index file
        blastIndexDirs: list of directories for trying to build index in
        '''
        self.seqDB = seqDB
        self.idIndex = BlastIDIndex(seqDB)
        self.verbose = verbose
        self.showFormatdbMessages = showFormatdbMessages
        if filepath is not None:
            self.filepath = filepath
        else:
            self.filepath = seqDB.filepath
        if blastIndexPath is not None:
            self.blastIndexPath = blastIndexPath
        if blastIndexDirs is not None:
            self.blastIndexDirs = blastIndexDirs
        self.checkdb() # CHECK WHETHER BLAST INDEX FILE IS PRESENT...
        if not self.blastReady and blastReady: # FORCE CONSTRUCTION OF BLAST DB
            self.formatdb()

    def __repr__(self):
        return "<BlastMapping '%s'>" % (self.filepath)

    def __getitem__(self, k):
        'return NLMSASlice representing BLAST results'
        al = self.__call__(k) # run BLAST & get NLMSA storing results
        return al[k] # return NLMSASlice representing these results

    def test_db_location(self, testpath):
        '''check whether BLAST index files ready for use; return status.'''
        if not os.access(testpath+'.nsd', os.R_OK) \
               and not os.access(testpath+'.psd', os.R_OK) \
               and not os.access(testpath+'.00.nsd', os.R_OK) \
               and not os.access(testpath+'.00.psd', os.R_OK):
            return False
        else: # FOUND INDEX FILES IN THIS LOCATION
            if testpath != self.filepath:
                self.blastIndexPath = testpath
            return True

    def checkdb(self):
        'look for blast index files in blastIndexPath, \
        standard list of locations'
        for testpath in self.blast_index_paths():
            self.blastReady = self.test_db_location(testpath)
            if self.blastReady:
                break
        return self.blastReady

    def run_formatdb(self, testpath):
        'ATTEMPT TO BUILD BLAST DATABASE INDEXES at testpath'
        dirname = classutil.file_dirpath(testpath)
        if not os.access(dirname, os.W_OK): # check if directory is writable
            raise IOError('run_formatdb: directory %s is not writable!'
                          % dirname)
        cmd = ['formatdb', '-i', self.filepath, '-n', testpath, '-o', 'T']
        if self.seqDB._seqtype != PROTEIN_SEQTYPE:
            cmd += ['-p', 'F'] # special flag required for nucleotide seqs
        logger.info('Building index: ' + ' '.join(cmd))
        if self.showFormatdbMessages:
            kwargs = {}
        else: # suppress formatdb messages by redirecting them
            kwargs = dict(stdout=classutil.PIPE, stderr=classutil.PIPE)
        if classutil.call_subprocess(cmd, **kwargs):
            # bad exit code, so command failed
            warn_if_whitespace(self.filepath) \
                 or warn_if_whitespace(testpath) # only issue one warning
            raise OSError('command %s failed' % ' '.join(cmd))
        self.blastReady=True
        if testpath!=self.filepath:
            self.blastIndexPath = testpath

    def get_blast_index_path(self):
        'get path to base name for BLAST index files'
        try:
            return self.blastIndexPath
        except AttributeError:
            return self.filepath
    # DEFAULT: BUILD INDEX FILES IN self.filepath . HOME OR APPROPRIATE
    # USER-/SYSTEM-SPECIFIC TEMPORARY DIRECTORY
    blastIndexDirs = ['FILEPATH', os.getcwd, os.path.expanduser,
                      tempfile.gettempdir()]

    def blast_index_paths(self):
        'iterate over possible blast index directories'
        try: # 1ST TRY ACTUAL SAVED LOCATION IF ANY
            yield self.blastIndexPath
        except AttributeError:
            pass
        for m in self.blastIndexDirs: # NOW TRY STANDARD LOCATIONS
            if m=='FILEPATH':
                yield self.filepath
                continue
            elif m == os.path.expanduser:
                s = m('~') # GET HOME DIRECTORY
            elif callable(m):
                s = m()
            else: # TREAT AS STRING
                s = str(m)
            yield os.path.join(s, os.path.basename(self.filepath))

    def formatdb(self, filepath=None):
        'try to build BLAST index files in an appropriate location'
        if filepath is not None: # JUST USE THE SPECIFIED PATH
            return self.run_formatdb(filepath)
        notFirst = False
        for testpath in self.blast_index_paths():
            if notFirst:
                logger.info('Trying next entry in self.blastIndexDirs...')
            notFirst = True
            try: # BUILD IN TARGET DIRECTORY
                return self.run_formatdb(testpath)
            except (IOError, OSError): # BUILD FAILED
                pass
        raise IOError("cannot build BLAST database for %s" % (self.filepath, ))

    def raw_fasta_stream(self, ifile=None, idFilter=None):
        '''Return a stream of fasta-formatted sequences.

        Optionally, apply an ID filter function if supplied.
        '''
        if ifile is not None: # JUST USE THE STREAM WE ALREADY HAVE OPEN
            return ifile, idFilter
        try: # DEFAULT: JUST READ THE FASTA FILE, IF IT EXISTS
            return file(self.filepath, 'rU'), idFilter
        except IOError: # TRY READING FROM FORMATTED BLAST DATABASE
            cmd='fastacmd -D -d "%s"' % self.get_blast_index_path()
            return os.popen(cmd), NCBI_ID_PARSER #BLAST ADDS lcl| TO id


    _blast_prog_dict = dict(blastx='#BlastxMapping')

    def blast_program(self, seq, blastprog=None):
        'figure out appropriate blast program & remap via _blast_prog_dict'
        if blastprog is None:
            blastprog = blast_program(seq.seqtype(), self.seqDB._seqtype)
        oldprog = blastprog
        try: # apply program transformation if provided
            blastprog = self._blast_prog_dict[blastprog]
            if blastprog.startswith('#'): # not permitted by this class!
                raise ValueError('Use %s for %s' % (blastprog[1:], oldprog))
        except KeyError:
            pass # no program transformation to apply, so nothing to do...
        return blastprog

    def blast_command(self, blastpath, blastprog, expmax, maxseq, opts):
        'generate command string for running blast with desired options'
        filepath = self.get_blast_index_path()
        warn_if_whitespace(filepath)
        cmd = [blastpath, '-d', filepath, '-p', blastprog,
                '-e', '%e' % float(expmax)] + list(opts)
        if maxseq is not None: # ONLY TAKE TOP maxseq HITS
            cmd += ['-b', '%d' % maxseq, '-v', '%d' % maxseq]
        return cmd

    def get_seq_from_queryDB(self, queryDB):
        'get one sequence obj from queryDB'
        seqID = iter(queryDB).next() # get 1st seq ID
        return queryDB[seqID]

    def translation_kwargs(self, blastprog):
        'return kwargs for read_blast_alignment() based on blastprog'
        d = dict(tblastn=dict(translateDest=True),
                 blastx=dict(translateSrc=True),
                 tblastx=dict(translateSrc=True, translateDest=True))
        try:
            return d[blastprog]
        except KeyError:
            return {}

    def __call__(self, seq=None, al=None, blastpath='blastall',
                 blastprog=None, expmax=0.001, maxseq=None, verbose=None,
                 opts=(), queryDB=None, **kwargs):
        "Run blast search for seq in database, return alignment object"
        if seq is None and queryDB is None:
            raise ValueError("we need a sequence or db to use as query!")
        if seq and queryDB:
            raise ValueError("both a sequence AND a db provided for query")
        if queryDB is not None:
            seq = self.get_seq_from_queryDB(queryDB)
        if not self.blastReady: # HAVE TO BUILD THE formatdb FILES...
            self.formatdb()
        blastprog = self.blast_program(seq, blastprog)
        cmd = self.blast_command(blastpath, blastprog, expmax, maxseq, opts)
        return process_blast(cmd, seq, self.idIndex, al, queryDB=queryDB,
                             ** self.translation_kwargs(blastprog))


class BlastxMapping(BlastMapping):
    """Because blastx changes the query to multiple sequences
    (representing its six possible frames), getitem can no longer
    return a single slice object, but instead an iterator for one
    or more slice objects representing the frames that had
    homology hits."""

    def __repr__(self):
        return "<BlastxMapping '%s'>" % (self.filepath)
    _blast_prog_dict = dict(blastn='tblastx', blastp='#BlastMapping',
                            tblastn='#BlastMapping')

    def __getitem__(self, query):
        """generate slices for all translations of the query """
        # generate NLMSA for this single sequence
        al = self(query)
        # get the translation database for the sequence
        tdb = translationDB.get_translation_db(query.db)

        # run through all of the frames & find alignments.
        slices = []
        for trans_seq in tdb[query.id].iter_frames():
            try:
                slice = al[trans_seq]
            except KeyError:
                continue

            if not isinstance(slice, EmptySlice):
                slices.append(slice)

        return slices


class MegablastMapping(BlastMapping):

    def __repr__(self):
        return "<MegablastMapping '%s'>" % (self.filepath)

    def __call__(self, seq, al=None, blastpath='megablast', expmax=1e-20,
                 maxseq=None, minIdentity=None,
                 maskOpts=['-U', 'T', '-F', 'm'],
                 rmPath='RepeatMasker', rmOpts=['-xsmall'],
                 verbose=None, opts=(), **kwargs):
        "Run megablast search with optional repeat masking."
        if not self.blastReady: # HAVE TO BUILD THE formatdb FILES...
            self.formatdb()

        # mask repeats to lowercase
        masked_seq = repeat_mask(seq, rmPath, rmOpts)
        filepath = self.get_blast_index_path()
        warn_if_whitespace(filepath)
        cmd = [blastpath] + maskOpts \
              + ['-d', filepath,
                 '-D', '2', '-e', '%e' % float(expmax)] + list(opts)
        if maxseq is not None: # ONLY TAKE TOP maxseq HITS
            cmd += ['-b', '%d' % maxseq, '-v', '%d' % maxseq]
        if minIdentity is not None:
            cmd += ['-p', '%f' % float(minIdentity)]
        return process_blast(cmd, seq, self.idIndex, al, seqString=masked_seq,
                             popenArgs=dict(stdinFlag='-i'))


class BlastIDIndex(object):
    """This class acts as a wrapper around a regular seqDB, and handles
    the mangled IDs returned by BLAST to translate them to the correct ID.
    Since NCBI treats FASTA ID as a blob into which they like to stuff
    many fields... and then NCBI BLAST mangles those IDs when it reports
    hits, so they no longer match the true ID... we are forced into
    contortions to rescue the true ID from mangled IDs.

    Our workaround strategy: since NCBI packs the FASTA ID with multiple
    IDs (GI, GB, RefSeq ID etc.), we can use any of these identifiers
    that are found in a mangled ID, by storing a mapping of these
    sub-identifiers to the true FASTA ID."""

    def __init__(self, seqDB):
        self.seqDB = seqDB
        self.seqInfoDict = BlastIDInfoDict(self)
    # FOR UNPACKING NCBI IDENTIFIERS AS WORKAROUND FOR BLAST ID CRAZINESS
    id_delimiter='|'

    def unpack_id(self, id):
        """Return |-packed NCBI identifiers as unpacked list.

        NCBI packs identifier like gi|123456|gb|A12345|other|nonsense.
        Return as list."""
        return id.split(self.id_delimiter)

    def index_unpacked_ids(self, unpack_f=None):
        """Build an index of sub-IDs (unpacked from NCBI nasty habit
        of using the FASTA ID as a blob); you can customize the unpacking
        by overriding the unpack_id function or changing the id_delimiter.
        The index maps each sub-ID to the real ID, so that when BLAST
        hands back a mangled, fragmentary ID, we can unpack that mangled ID
        and look up the true ID in this index.  Any sub-ID that is found
        to map to more than one true ID will be mapped to None (so that
        random NCBI garbage like gnl or NCBI_MITO wont be treated as
        sub-IDs).
        """
        if unpack_f is None:
            unpack_f=self.unpack_id
        t={}
        for id in self.seqDB:
            for s in unpack_f(id):
                if s == id:
                    continue # DON'T STORE TRIVIAL MAPPINGS!!
                s=s.upper() # NCBI FORCES ID TO UPPERCASE?!?!
                try:
                    if t[s]!=id and t[s] is not None:
                        t[s]=None # s NOT UNIQUE, CAN'T BE AN IDENTIFIER!!
                except KeyError:
                    t[s]=id # s UNIQUE, TRY USING s AS AN IDENTIFIER
        for id in t.itervalues():
            if id is not None: # OK THERE ARE REAL MAPPINGS STORED, SO USE THIS
                self._unpacked_dict=t # SAVE THE MAPPING TO REAL IDENTIFIERS
                return
        # NO NON-TRIVIAL MAPPINGS, SO JUST SAVE EMPTY MAPPING
        self._unpacked_dict={}

    def get_real_id(self, bogusID, unpack_f=None):
        "try to translate an id that NCBI has mangled to the real sequence id"
        if unpack_f is None:
            unpack_f = self.unpack_id
        if not hasattr(self, '_unpacked_dict'):
            self.index_unpacked_ids(unpack_f)
        for s in unpack_f(bogusID):
            s = s.upper() # NCBI FORCES ID TO UPPERCASE?!?!
            try:
                id = self._unpacked_dict[s]
                if id is not None:
                    return id # OK, FOUND A MAPPING TO REAL ID
            except KeyError:
                pass # KEEP TRYING...
        # FOUND NO MAPPING, SO RAISE EXCEPTION
        raise KeyError("no key '%s' in database %s" % (bogusID,
                                                        repr(self.seqDB)))

    def __getitem__(self, seqID):
        "If seqID is mangled by BLAST, use our index to get correct ID"
        try: # default: treat as a correct ID
            return self.seqDB[seqID]
        except KeyError: # translate to the correct ID
            return self.seqDB[self.get_real_id(seqID)]

    def __contains__(self, seqID):
        try:
            self.seqInfoDict[seqID]
            return True
        except KeyError:
            return False


class BlastIDInfoDict(object, UserDict.DictMixin):
    """provide seqInfoDict interface for BlastIDIndex """

    def __init__(self, db):
        self.blastDB = db

    def __getitem__(self, seqID):
        try:
            return self.blastDB.seqDB.seqInfoDict[seqID]
        except KeyError:
            seqID = self.blastDB.get_real_id(seqID)
            return self.blastDB.seqDB.seqInfoDict[seqID]

    def __len__(self):
        return len(self.blastDB.seqDB.seqInfoDict)

    def __iter__(self):
        return iter(self.blastDB.seqDB.seqInfoDict)

    def keys(self):
        return self.blastDB.seqDB.seqInfoDict.keys()



from __future__ import generators
from pygr.sequence import *
#from pathquery import *


class AnonSequence(Sequence):
    """Defines a sequence class with unknown sequence, but
    known length"""

    def __init__(self, length, id):
        s = ''
        self.known = list()
        Sequence.__init__(self, s, id)
        self.stop = length

    def seqsplice(self, s, start, end):
        (begin, stop, step) = slice(start, end).indices(self.stop)

        if (start < end):
            self.known += [(s, start, stop)]
        elif start > end:
            self.known += [(s[::-1], start, stop)]

    def known_int(self):
        for u in self.known:
            yield {'src_id': self.id, 'start': u[1], 'end': u[2], 'seq': u[0]}


class ReferenceSequence(Sequence):
    """Defines a reference sequence class that is subscriptable
    by other sequences. If sequence ids match the resulting sequnce
    will reference this class. This is useful for coordinate
    transforms when an unknown sequence intervals are transformed
    to known sequence"""

    def __init__(self, s, id):
        Sequence.__init__(self, s, id)

    def __getitem__(self, iv):
        if(isinstance(iv, SeqPath)):
            if(iv.id == self.id):
                s = self[iv.start:iv.stop:iv.step]
                s.orientation = iv.orientation
                return s
        else:
            return SeqPath.__getitem__(self, iv)

    def mapCoordinates(self, obj):
        m = PathMapping2()
        for ival in obj:
            m[ival] = self[ival]
        return m


class UnkSequence(SeqPath):
    """Defines a sequence class for pure interval manipulation.
    No sequence information (i.e. length, or seq itself) is needed.
    """

    def __init__(self, id, start=0, end=0, step=1, orientation=1):
        self.id = id
        SeqPath.__init__(self, id)
        self.start = start
        self.stop = end
        self.step = step
        self.orientation = orientation
        if (self.start is not None and self.stop is not None
           and self.start > self.stop):
            t = self.start
            if (self.stop >= 0):
                self.start = self.stop + 1
            else:
                self.start = self.stop
            if (t >= 0):
                self.stop = t + 1
            else:
                self.stop = t
            self.orientation = -self.orientation

    def __getitem__(self, k):
        if isinstance(k, types.SliceType):
            (start, stop, step) = (k.start, k.stop, k.step)
            if k.step == None:
                step = 1
        elif isinstance(k, types.IntType):
            start = k
            stop = k + 1
            if (k == -1):
                stop = None
            step = 1
            return self[start:stop:step]
        else:
            raise KeyError('requires a slice object or integer key')
        if self.step == 1 and stop != None:
            return UnkSequence(self.id, self.start + start * self.step,
                           self.start + stop * self.step,
                           self.step * step, self.orientation)
        else:
            return UnkSequence(self.id, start, stop, step, self.orientation)




from __future__ import generators
import types
from sequtil import *


NOT_ON_SAME_PATH = -2


class ReadOnlyAttribute(object):

    def __init__(self, attr):
        self.attr = attr

    def __get__(self, obj, klass):
        return getattr(obj, self.attr)


###################################################################
#
# INTERVAL - INTERVAL MAPPING
# CORRECTLY HANDLES SCALING AND ORIENTATION TRANSFORMATIONS
# ASSUMES A SIMPLE SCALAR TRANSFORMATION FROM ONE COORD SYSTEM
# TO THE OTHER.  DOES NOT HANDLE INDELS WITHIN THE MAPPING!

class IntervalTransform(object):
    "Represents coordinate transformation from one interval to another"
    srcPath=ReadOnlyAttribute('_srcPath') # PREVENT USER FROM MODIFYING THESE!
    destPath=ReadOnlyAttribute('_destPath')

    def __init__(self, srcPath, destPath, edgeInfo=None,
                 edgeAttr=None, edgeIndex=None):
        "MAP FROM srcPath -> destPath"
        self.scale = len(destPath) / float(len(srcPath))
        self._srcPath = srcPath
        self._destPath = destPath
        if edgeInfo != None and edgeAttr != None:
            try: # GET EDGE INFO IF PRESENT
                edgeInfo = getattr(edgeInfo, edgeAttr)
            except AttributeError:
                edgeInfo = None
        if edgeInfo != None:
            if edgeIndex != None:
                edgeInfo = edgeInfo[edgeIndex]
            self.edgeInfo = edgeInfo

    def xform(self, i):
        'transform int srcPath local coord to destPath local coord'
        return int(i * self.scale)

    def xformBack(self, i):
        'transform int destPath local coord to srcPath local coord'
        return int(i / self.scale)

    def getStartStop(self, srcPath, ourPath):
        'compute srcPath start,stop in ourPath local coords'
        if srcPath.path is ourPath.path:
            return srcPath.start - ourPath.start, srcPath.stop - ourPath.start
        try:
            if srcPath.path._reverse is ourPath.path:
                return -(srcPath.start) - ourPath.start,\
                       -(srcPath.stop) - ourPath.start
        except AttributeError:
            pass
        raise ValueError('sequence mismatch: argument not from this seq')

    def __call__(self, srcPath):
        """Apply this transformation to an interval
           NB: it is not restricted to the domain of this transform,
           and thus can extend BEYOND the boundaries of this transform.
           If you want it clipped use [] interface instead of ()."""
        start, stop = self.getStartStop(srcPath, self.srcPath)
        return SeqPath(self.destPath, self.xform(start), self.xform(stop),
                       relativeToStart=True)

    def reverse(self, destPath):
        "reverse transform an interval"
        start, stop = self.getStartStop(destPath, self.destPath)
        return SeqPath(self.srcPath, self.xformBack(start),
                       self.xformBack(stop), relativeToStart=True)

    def __getitem__(self, srcPath): # PROVIDE DICT-LIKE INTERFACE
        """intersect srcPath with domain of this transform, then return
        transform to target domain coordinates"""
        return self(srcPath * self.srcPath)

    def __iter__(self):
        yield self.srcPath

    def items(self):
        yield self.srcPath, self.destPath

    def __getattr__(self, attr):
        "provide transparent wrapper for edgeInfo attributes"
        try:
            return getattr(self.__dict__['edgeInfo'], attr)
        except (KeyError, AttributeError): # RAISE EXCEPTION IF NOT FOUND!
            raise AttributeError('neither IntervalTransform nor edgeinfo \
                                 has attr ' + attr)

    def repr_dict(self):
        "Return compact dictionary representing this interval mapping"
        s = self.srcPath.repr_dict() # GET REPR OF BOTH INTERVALS
        d = self.destPath.repr_dict()
        out= {}
        for k, val in s.items(): # ADD PREFIX TO EACH ATTR
            out['src_' + k] = val
            out['dest_' + k] = d[k]
        try:
            e = self.edgeInfo.repr_dict() # GET EDGE INFO IF PRESENT
        except AttributeError:
            pass
        else:
            out.update(e) # SAVE EDGE INFO DATA
        return out

    def nidentity(self):
        "calculate total #identity matches between srcPath and destPath"
        nid = 0
        src = str(self.srcPath).upper()
        dest = str(self.destPath).upper()
        slen = len(src)
        i = 0
        while i < slen:
            if src[i] == dest[i]:
                nid += 1
            i += 1
        return nid

    def percent_id(self):
        "calculate fractional identity for this pairwise alignment"
        return self.nidentity / float(len(self.srcPath))


###################################################################
#
# SINGLE LETTER GRAPH INTERFACE CLASSES
# INSTANTIATED ON THE FLY TO PROVIDE LETTER GRAPH INTERFACE

class LetterEdgeDescriptor(object):
    "cached dict of sequences traversing this edge"

    def __get__(self, edge, objtype):
        try:
            return edge._seqs
        except AttributeError:
            edge._seqs = edge.origin.getEdgeSeqs(edge.target)
            return edge._seqs


class LetterEdge(object):
    "represents an edge from origin -> target. seqs returns its sequences"
    seqs = LetterEdgeDescriptor()

    def __init__(self, origin, target):
        self.origin = origin
        self.target = target

    def __iter__(self):
        'generate origin seqpos for sequences that traverse this edge'
        for seq in self.seqs:
            # Reduce 1:many schema to 1:1, discard edge information.
            yield self.origin.getSeqPos(seq)

    def iteritems(self):
        'generate origin,target seqpos for sequences that traverse this edge'
        for seq in self.seqs: # REDUCE 1:MANY SCHEMA TO 1:1, DISCARD EDGE INFO
            yield self.origin.getSeqPos(seq), self.target.getSeqPos(seq)

    def __getitem__(self, k):
        '''return origin,target seqpos for sequence k; raise KeyError if
        not in this edge'''
        try:
            k = k.path
        except AttributeError:
            raise TypeError('k not a valid sequence: it has no path attribute')
        return (self.origin.getSeqPos(k.path), self.target.getSeqPos(k.path))

    def __cmp__(self, other):
        'two edge objects match if they link identical nodes'
        try:
            if self.origin == other.origin and self.target == other.target:
                return 0 # REPORT A MATCH
        except AttributeError: # other MUST BE A DIFFERENT TYPE, SO NO MATCH
            pass
        return cmp(id(self), id(other))


def absoluteSlice(seq, start, stop):
    '''get slice of top-level sequence object, in absolute coordinates.
    This method calls getitem on the top-level sequence object
    i.e. seq.pathForward'''
    try:
        if start < 0: # REVERSE ORIENTATION
            return -(seq.pathForward[-stop:-start])
        else: # FORWARD ORIENTATION
            return seq.pathForward[start:stop]
    except AttributeError:
        if seq is None:
            return slice(start, stop)


def relativeSlice(seq, start, stop):
    '''get slice of this sequence object, in relative coordinates.
    This method calls getitem on the top-level sequence object
    i.e. seq.pathForward'''
    if start < 0: # REVERSE ORIENTATION
        return -(seq[-stop: -start])
    else: # FORWARD ORIENTATION
        return seq[start: stop]


def sumSliceIndex(i, myslice, relativeToStart):
    '''Adjust index value either relative to myslice.start (positive indexes)
    or relative to myslice.stop (negative indexes).  Handle the case where
    i is None or myslice is None appropriately.
    '''
    if myslice is None: # NO OBJECT, SO NOTHING TO DO...
        return i
    if i is None:
        i = 0
    i *= myslice.step
    if relativeToStart:
        return i + myslice.start
    else:
        return i + myslice.stop


class SeqOriDescriptor(object):
    "Get orientation of sequence interval"

    def __get__(self, seq, objtype):
        if seq.start >= 0:
            return 1 # FORWARD ORIENTATION
        else:
            return -1 # REVERSE ORIENTATION


class PathForwardDescr(object):
    'get the top-level forward sequence object'

    def __get__(self, seq, objtype):
        if seq.orientation > 0:
            return seq.path
        else:
            return seq.path._reverse


class AbsIntervalDescr(object):
    'get the top-level forward sequence object'

    def __get__(self, seq, objtype):
        if seq.orientation > 0:
            return seq.start, seq.stop
        else:
            return -(seq.stop), -(seq.start)


class SeqPath(object):
    '''Base class for specifying a path, ie. sequence interval.
    This implementation takes a sequence object as initializer
    and simply represents the interval as a slice of the sequence.'''
    orientation=SeqOriDescriptor()  # COMPUTE ORIENTATION AUTOMATICALLY
    pathForward=PathForwardDescr()  # GET THE TOP-LEVEL FORWARD SEQUENCE OBJ
    _abs_interval=AbsIntervalDescr()

    def __init__(self, path, start=0, stop=None, step=None, reversePath=False,
                 relativeToStart=False, absoluteCoords=False):
        '''Return slice of path[start:stop:step].
        NB: start>stop means reverse orientation, i.e. (-path)[-stop:-start]
        start/stop are LOCAL coordinates relative to the specified path
        By default, start/stop are interpreted in the usual Python slicing way,
        i.e. a negative start value is interpreted as len(path)-start.
        The relativeToStart option turns off this behavior, so that negative
        values are interpreted as negative coordinates in the local coordinate
        system of path.

        absoluteCoords option allows intervals to be created using
        Pygr's internal coordinate convention i.e.
        -20,-10 --> -(path.pathForward[10:20])
        '''
        if reversePath: # create top-level negative orientation path
            start = -(path.stop)
            stop = 0
            self._reverse = path
            path = None # make this a top-level path object
        if absoluteCoords: # THIS OPTION PROVIDES TRANSPARENT WAY TO CREATE
            if start >= 0:   # INTERVALS USING start,stop PAIRS THAT FOLLOW
                path = path.pathForward # OUR INTERNAL SIGN CONVENTION
            else: # i.e. start<0 MEANS REVERSE ORIENTATION!
                path = - (path.pathForward)
        else: # ADJUST start,stop ACCORDING TO path.start / path.stop
            start = sumSliceIndex(start, path, relativeToStart or start is None
                                  or start>=0)
            stop = sumSliceIndex(stop, path, relativeToStart or
                                 (stop is not None and stop>=0))
        if start is not None and stop is not None and start>stop:
            start = -start # start>stop MEANS REVERSE ORIENTATION!
            stop = -stop
            if path is not None:
                path = -path
        if path is not None: # perform bounds checking
            if start < path.path.start:
                start = path.path.start
            if stop > path.path.stop:
                stop = path.path.stop
        if start is None or stop is None or start>=stop:
            raise IndexError('cannot create empty sequence interval!')
        self.start = start
        self.stop = stop
        if step is None:
            step = 1
        if path is None:
            self.path = self
            self.step = step
        else: # STORE TOP-LEVEL SEQUENCE PATH...
            self.path = path.path
            self.step = step * path.step

    def classySlice(self, path, *l, **kwargs):
        'create a subslice using appropriate class based on container'
        try: # if db provides a class to use for slices, use it.
            klass = path.pathForward.db.itemSliceClass
        except AttributeError:
            klass = SeqPath # default: just use generic slice class
        return klass(path, *l, **kwargs) # construct the slice

    def absolute_slice(self, start, stop):
        'get slice of top-level sequence, using absolute coords'
        return self.classySlice(self, start, stop, absoluteCoords=True)

    def __getitem__(self, k):
        if isinstance(k, types.IntType):
            if k == -1: # HAVE TO HANDLE THIS CASE SPECIALLY
                k = slice(k, None, 1) # -1 IS LAST LETTER, SO [-1:None] slice
            else: # REGULAR CASE, JUST [k:k+1] slice
                k = slice(k, k + 1, 1)
        if isinstance(k, types.SliceType): # GET AN INTERVAL USING slice
            return self.classySlice(self, k.start, k.stop, k.step)
        elif isinstance(k, SeqPath): # MODEL SEQ AS GRAPH
            if k.path is not self.path:
                raise KeyError('node is not in this sequence!')
            try:
                target = self.classySlice(self.path, k.stop,
                                          k.stop + len(k) * k.step, k.step)
                return {target: LetterEdge(k, target)}
            except IndexError: # OUT OF BOUNDS, SO NO NEXT NODE
                return {}
        raise KeyError('requires a slice object or integer key')

    def __len__(self):
        if self.path is self and self.orientation < 0:
            return len(self._reverse) # GET LENGTH FROM FORWARD SEQUENCE
        # Get the number of results from iter(self).
        d = (self.stop - self.start) / self.step
        if d > 0: # IF stop - start < step, d WILL BE ZERO -- PREVENT THAT!
            return d
        else:
            # Never return 0 length... Bounds checking ensures non-empty ival.
            return 1

    ################################ LETTER GRAPH METHODS: JUST A LINEAR GRAPH
    def __iter__(self):
        for i in range(len(self)):
            yield self[i]

    def iteritems(self):
        'letter graph iterator over (node1, {node2: edge}) tuples'
        src = self[0] # GET 1ST LETTER
        for i in range(1, len(self)): # ITERATE OVER ALL ADJACENT PAIRS
            target = self[i]
            yield src, {target: LetterEdge(src, target)}
            src = target
        yield src, {} # LAST LETTER HAS NO EDGES

    def getEdgeSeqs(self, other):
        'return dict of sequences that traverse edge from self -> other'
        if self.path is other.path and self.stop == other.start:
            return {self.path: self.stop - self.step}
        else:
            return {}

    def getSeqPos(self, seq):
        'get seq interval corresponding to this node in sequence graph'
        if seq.path is self.path:
            return self
        raise KeyError('seq not on this path!')

    ################################ INTERVAL COMPOSITION OPERATORS
    def __hash__(self):
        'ensure that same seq intervals match in dict'
        return id(self.path)^hash(self.start)^hash(self.stop)

    def __cmp__(self, other):
        'ensure that same seq intervals match in cmp()'
        if not isinstance(other, SeqPath):
            return -1
        if self.path is other.path:
            return cmp((self.start, self.stop), (other.start, other.stop))
        else:
            return NOT_ON_SAME_PATH
            #raise TypeError('SeqPath not comparable, not on same path: %s, %s'
            #                % (self.path, other.path))

    def __contains__(self, k):
        # PUT OTHER LOGIC HERE FOR CHECKING WHETHER INTERVAL IS CONTAINED...
        if isinstance(k, SeqPath):
            if k.path == self.path and self.start <= k.start and \
               k.stop <= self.stop:
                return True
            else:
                return False
        elif isinstance(k, types.IntType):
            return self.start <= k and k < self.stop

    def overlaps(self, p):
        "check whether two paths on same seq overlap"
        if self.path is not p.path:
            return False
        if (self.start <= p.start and p.start < self.stop) or \
               (p.start <= self.start and self.start < p.stop):
            return True
        else:
            return False

    def __mul__(self, other):
        "find intersection of two intervals"
        if isinstance(other, SeqPath):
            if self.path is not other.path:
                return None
            start=max(self.start, other.start)
            stop=min(self.stop, other.stop)
            if start < stop:
                if stop == 0:
                    # Have to handle the boundary case specially because of
                    # Python conventions.
                    stop = None
                return self.classySlice(self.path, start, stop)
            else:
                return None
        else:
            raise TypeError('SeqPath can only intersect SeqPath')

    def __div__(self, other):
        "return transform from other -> self coordinate systems"
        return IntervalTransform(other, self)

    def __neg__(self):
        "return same interval in reverse orientation"
        try:
            if self.seqtype() == PROTEIN_SEQTYPE:
                raise ValueError('protein sequence has no reverse \
                                 orientation!')
        except AttributeError:
            pass # ALLOW UNTYPED SEQ OBJECTS TO BE REV-COMPD
        if self is self.path: # TOP-LEVEL SEQUENCE OBJECT
            try:
                return self._reverse # USE EXISTING RC OBJECT FOR THIS SEQ
            except AttributeError: #  CREATE ONLY ONE RC FOR THIS SEQUENCE
                self._reverse = self.classySlice(self, reversePath=True)
                return self._reverse
        elif self.orientation > 0: # FORWARD ORI: JUST REVERSE INDICES
            return self.classySlice(self.path, self.stop, self.start,
                                    self.step) #SWAP ==> RC
        else: # REVERSE ORI: BECAUSE OF stop=0 POSSIBILITY, USE POSITIVE COORDS
            return self.classySlice(self.path._reverse, -(self.stop),
                                    -(self.start), self.step)

    def __add__(self, other):
        "return merged interval spanning both self and other intervals"
        if self.path is not other.path:
            raise ValueError('incompatible intervals cannot be merged.')
        if self.start < other.start:
            start = self.start
        else:
            start = other.start
        if self.stop > other.stop:
            stop = self.stop
        else:
            stop = other.stop
        if stop == 0:
            # Have to handle the boundary case specially because of Python
            # conventions.
            stop = None
        return self.classySlice(self.path, start, stop, self.step)

    def __iadd__(self, other):
        "return merged interval spanning both self and other intervals"
        if self.path is not other.path:
            raise ValueError('incompatible intervals cannot be merged.')
        if other.start < self.start:
            self.start = other.start
        if other.stop > self.stop:
            self.stop = other.stop
        return self # iadd MUST ALWAYS RETURN self!!

    def before(self):
        'get the sequence interval before this interval'
        return self.classySlice(self.path, None, self.start)

    def after(self):
        'get the sequence interval after this interval'
        if self.stop == 0:
            raise IndexError('cannot create empty sequence interval')
        return self.classySlice(self.path, self.stop, None)

    def is_full_length(self):
        '''test whether this constitutes the whole sequence (in either
        orientation)'''
        return self == self.path

    ############################################ STRING SEQUENCE METHODS
    _complement = {'a': 't', 'c': 'g', 'g': 'c', 't': 'a', 'u': 'a', 'n': 'n',
                   'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A', 'U': 'A', 'N': 'N'}

    def reverse_complement(self, s):
        'get reverse complement of a string s'
        #return ''.join([self._complement.get(c, c) for c in s[::-1]])
        return ''.join([self._complement.get(s[i], s[i]) for i in
                        range(len(s) -1, -1, -1)])

    def seqtype(self):
        "Get the sequence type for this sequence"
        path = self.pathForward
        try: # TRY GETTING IT FROM TOP-LEVEL SEQUENCE OBJECT?
            return path._seqtype
        except AttributeError:
            try: # TRY TO GET IT FROM DB THIS SEQ IS ASSOCIATED WITH, IF ANY
                return path.db._seqtype
            except AttributeError:# GUESS IT FROM 1ST 40 LETTERS OF SEQUENCE
                path._seqtype = guess_seqtype(str(self[0:40]))
                return path._seqtype

    def __str__(self):
        '''string for this sequence interval; use reverse complement
        if necessary...'''
        if self.orientation > 0:
            return self.path.strslice(self.start, self.stop)
        else:
            s = self.path._reverse.strslice(-(self.stop), -(self.start))
            return self.reverse_complement(s)

    def __repr__(self):
        try: # USE id CONVENTION TO GET A NAME FOR THIS SEQUENCE
            id = self.pathForward.id
        except AttributeError: # OTHERWISE USE A DEFAULT, SHOWING THERE'S NO id
            id = '@NONAME'
        if self.orientation < 0: # INDICATE NEGATIVE ORIENTATION
            return '-%s[%d:%d]' % (id, -self.stop, -self.start)
        else:
            return '%s[%d:%d]' % (id, self.start, self.stop)

    def repr_dict(self):
        "Return compact dictionary representing this interval"
        try:
            id = self.path.id
        except AttributeError:
            id = self.id
        return {'id': id, 'start': self.start, 'end': self.stop,
                'ori': self.orientation}


class LengthDescriptor(object):
    'property that just returns length of object'

    def __get__(self, seq, klass):
        return len(seq)


# BASIC WRAPPER FOR A SEQUENCE.  LETS US ATTACH A NAME TO IT...
class SequenceBase(SeqPath):
    '''base sequence type assumes there will be a seq attribute
    providing sequence'''
    start = 0
    stop = LengthDescriptor()
    step = 1
    orientation = 1

    def __init__(self):
        self.path = self

    def update(self, seq):
        'change this sequence to the string <seq>'
        self.seq = seq

    def __len__(self):
        'default: get the whole self.seq and compute its length'
        return len(self.seq) # COMPUTE IT FROM THE SEQUENCE

    def strslice(self, start, stop):
        'default method assumes self.seq is a sliceable string'
        return self.seq[start:stop]


class Sequence(SequenceBase):
    'default sequence class initialized with a sequence string and ID'

    def __init__(self, s, id):
        SequenceBase.__init__(self)
        self.id = id
        self.seq = s


class SeqFilterDict(dict):
    '''stores a set of intervals, either on init or via self[ival]=junk;
    self[ival] returns intersection of ival and the overlapping
    interval in self if any; otherwise raise KeyError'''

    def __init__(self, l=[]):
        'accepts optional arg giving list of intervals'
        dict.__init__(self)
        for ival in l:
            self[ival] = None

    def __getitem__(self, k):
        try:
            ival = dict.__getitem__(self, k.path)
        except KeyError:
            raise KeyError('seq not in dict')
        result = k * ival # RETURN INTERSECTION OF IVALS
        if result is None:
            raise KeyError # PROPER WAY TO SIGNAL KEY MAPS TO NO VALUE
        return result

    def __setitem__(self, ival, junk):
        dict.__setitem__(self, ival.path, ival)

    def __iter__(self):
        return dict.itervalues(self)
##
##
## class S2SEEdgesDescriptor(object):
##     "list of interval matches as list of tuples (ival1, ival2, xform)"
##     def __get__(self, s2se, objtype):
##         return [(t.srcPath, t.destPath, t) for t in s2se.matches]


class Seq2SeqEdge(object):
    '''Maps two sequence regions onto each other, using a list
    of scalar transformations.  Can handle indels within the
    mapping.'''
    #edges = S2SEEdgesDescriptor()

    def __init__(self, msaSlice, targetPath, sourcePath=None,
                 matchIntervals=False):
        self.msaSlice = msaSlice
        self.targetPath = targetPath
        if sourcePath is not None:
            self.sourcePath = sourcePath
            self.matchIntervals = matchIntervals
        else: # NEED TO REVERSE-MAP targetPath TO FIND sourcePath
            # Mask to targetPath.
            si = msaSlice.groupByIntervals(filterList = [targetPath],
                                           mergeAll=True)
            l = msaSlice.groupBySequences(si)
            try:
                self.sourcePath = l[0][0]
            except IndexError:
                raise KeyError('target interval not in msaSlice!')
            self.matchIntervals = l[0][2]

    def items(self, mergeAll=False, **kwargs):
        'get list of (srcPath, destPath) 1:1 matches'
        if self.matchIntervals is None: # THIS IS ALREADY A 1:1 INTERVAL!
            return [(self.sourcePath, self.targetPath)]
        elif self.matchIntervals is False:
            raise ValueError('no matchIntervals information!')
        l = [] # USE STORED LIST OF 1:1 INTERVALS
        for srcStart, srcEnd, destStart, destEnd in self.matchIntervals:
            l.append((absoluteSlice(self.sourcePath, srcStart, srcEnd),
                      absoluteSlice(self.targetPath, destStart, destEnd)))
        return l

    def get_gaps(self):
        'return list of (srcIval,destIval) representing gaps / insertions'
        if self.matchIntervals is False:
            raise ValueError('no matchIntervals information!')
        elif self.matchIntervals is None or len(self.matchIntervals) < 2:
            return [] # no gaps here...
        srcLast = self.matchIntervals[0][1] # ends of 1st aligned intervals
        destLast = self.matchIntervals[0][3]
        l = []
        for t in self.matchIntervals[1:]:
            if t[0] > srcLast: # gap region
                srcIval = absoluteSlice(self.sourcePath, srcLast, t[0])
            else:
                srcIval = None
            if t[2] > destLast: # insertion region
                destIval = absoluteSlice(self.targetPath, destLast, t[2])
            else:
                destIval = None
            if srcIval is not None or destIval is not None:
                l.append((srcIval, destIval))
            srcLast = t[1] # ends of these aligned intervals
            destLast = t[3]
        return l

    def __iter__(self, sourceOnly=True, **kwargs):
        return iter([t[0] for t in self.items(sourceOnly=sourceOnly,
                                              **kwargs)])

    def length(self, mode=max):
        'get length of source vs. target interval according to mode'
        return mode(len(self.sourcePath), len(self.targetPath))

    def pIdentity(self, mode=max, trapOverflow=True):
        "calculate fractional identity for this pairwise alignment"
        nid = 0
        start1 = self.sourcePath.start
        s1 = str(self.sourcePath).upper()
        start2 = self.targetPath.start
        s2 = str(self.targetPath).upper()
        for srcPath, destPath in self.items():
            isrc = srcPath.start - start1
            idest = destPath.start - start2
            for i in xrange(len(srcPath)):
                if s1[isrc + i] == s2[idest + i]:
                    nid += 1
        x = nid / float(self.length(mode))
        if trapOverflow and x > 1.:
            raise ValueError('''pIdentity overflow due to multiple hits
                             (see docs)? To avoid this error message,
                             use trapOverflow=False option.''')
        return x

    def longestSegment(self, segment, pIdentityMin=.9, minAlignSize=1,
                       mode=max, **kwargs):
        besthit = None
        for i in xrange(len(segment)):
            ni = 0 # IDENTITY COUNT
            nm = 0 # MISMATCH COUNT
            for j in xrange(i, -1, -1):
                ni += segment[j][2]
                l = mode(segment[i][0] + segment[i][2] - segment[j][0],
                       segment[i][1] + segment[i][2] - segment[j][1])
                pIdentity = float(ni) / l
                if pIdentity >= pIdentityMin and (besthit is None or
                                                  ni + nm > besthit[4]):
                    besthit = (segment[j][0], segment[i][0] + segment[i][2],
                               segment[j][1], segment[i][1] + segment[i][2],
                               ni + nm)
                nm += segment[j][3]
        if besthit is None:
            return None
        elif besthit[4] >= minAlignSize:
            return besthit[:4]
        else:
            return None

    def conservedSegment(self, **kwargs):
        "calculate fractional identity for this pairwise alignment"
        start1 = self.sourcePath.start
        s1 = str(self.sourcePath).upper()
        start2 = self.targetPath.start
        s2 = str(self.targetPath).upper()
        i1 = None
        segment = []
        n = 0
        # Find unbroken identity segments.
        for srcPath, destPath in self.items():
            isrc = srcPath.start - start1
            idest = destPath.start - start2
            for i in xrange(len(srcPath)):
                if s1[isrc + i] == s2[idest + i]: # EXACT MATCH
                    if i1 is None: # START NEW IDENTITY-SEGMENT
                        seg1, i1 = isrc + i, isrc + i
                        seg2, i2 = idest + i, idest + i
                    elif i1 + 1 != isrc + i or i2 + 1 != idest + i:
                        # Not contiguous, break.
                        segment.append((seg1 + start1, seg2 + start2,
                                        i1 + 1 - seg1, n))
                        n = 0 # RESET MISMATCH COUNT
                        seg1, i1 = isrc + i, isrc + i
                        seg2, i2 = idest + i, idest + i
                    else:
                        i1 = isrc + i
                        i2 = idest + i
                else: # MISMATCH
                    if i1 is not None: # BREAK PREVIOUS SEGMENT
                        segment.append((seg1 + start1, seg2 + start2,
                                        i1 + 1 - seg1, n))
                        i1 = None
                        n = 0 # RESET MISMATCH COUNT
                    n += 1 # COUNT MISMATCH
        if i1 is not None:
            segment.append((seg1 + start1, seg2 + start2, i1 + 1 - seg1, n))
        return self.longestSegment(segment, **kwargs)

    def pAligned(self, mode=max, trapOverflow=True):
        'get fraction of aligned letters for this pairwise alignment'
        nid = 0
        for srcPath, destPath in self.items():
            nid += len(destPath)
        x = nid / float(self.length(mode))
        if trapOverflow and x > 1.:
            raise ValueError('''pAligned overflow due to multiple hits
                             (see docs)? To avoid this error message,
                             use trapOverflow=False option.''')
        return x


class SeqDBDescriptor(object):
    'forwards attribute requests to self.pathForward'

    def __init__(self, attr):
        self.attr = attr

    def __get__(self, obj, objtype):
        # Raises an AttributeError if None.
        return getattr(obj.pathForward, self.attr)


class SeqDBSlice(SeqPath):
    'JUST A WRAPPER FOR SCHEMA TO HANG SHADOW ATTRIBUTES ON...'
    id = SeqDBDescriptor('id')
    db = SeqDBDescriptor('db')



# CURRENTLY UNUSED

## class PathEdgeDict(dict):
##
##     def __init__(self, p):
##         self.path = p.path
##         self.pos = p.end - 1
##         if p.end < len(p.path):
##             dict.__setitem__(self, p.path[p.end], 1)
##         if hasattr(p.path, '_next') and self.pos in p.path._next:
##             dict.update(self, p.path._next[self.pos])
##
##     def __setitem__(self, k, val):
##         print 'entered PathEdgeDict.setitem'
##         if not hasattr(self.path, '_next'):
##             self.path._next = {}
##         if self.pos not in self.path._next:
##             self.path._next[self.pos] = {}
##         self.path._next[self.pos][k] = val
##         dict.__setitem__(self, k, val)


## class PathNextDescr(object):
##
##     def __init__(self, attrName='next'):
##         self.attrName = attrName
##
##     def __get__(self, obj, objtype):
##         return PathEdgeDict(obj)
##
##     def __set__(self, obj, val):
##         raise AttributeError(self.attrName + ' is read-only!')


## class LengthDescriptor(object):
##
##     def __init__(self, attr):
##         self.attr = attr
##
##     def __get__(self, obj, objtype):
##         return len(getattr(obj, self.attr))
##
##     def __set__(self, obj, val):
##         raise AttributeError(self.attr + ' is read-only!')


## def firstItem(aList):
##     if hasattr(aList, '__iter__'):
##         for i in aList:
##             return i
##     else:
##         return aList



import sys
__version__ = "0.8.1"

try:
    worldbase
except NameError:
    import metabase
    mdb = metabase.MetabaseList() # use default WORLDBASEPATH
    worldbase = mdb.Data

if sys.version_info < (2, 6):

    def dir(obj):
        """provide Python 2.6+ style __dir__ support """
        try:
            get_dir = obj.__dir__
        except AttributeError:
            return __builtins__['dir'](obj)
        else:
            return get_dir()
else: # Python 2.6 or later, just use the builtin dir()
    dir = __builtins__['dir']

__all__ = ('__version__', 'worldbase', 'dir')



import MySQLdb
import os
from splicegraph import *


spliceCalcs={'HUMAN_SPLICE_03':
             TableGroup(db='HUMAN_SPLICE_03', suffix='JUN03',
                        clusters='cluster_JUN03', exons='exon_formJUN03',
                        splices='splice_verification_JUN03',
                        genomic='genomic_cluster_JUN03', mrna='mrna_seqJUN03',
                        protein='protein_seqJUN03'),
             'HUMAN_SPLICE':
             TableGroup(db='HUMAN_SPLICE', suffix='jan02',
                        clusters='cluster_jan02',
                        exons='HUMAN_ISOFORMS.exon_form_4',
                        splices='splice_verification_jan02',
                        genomic='genomic_cluster_jan02',
                        mrna='HUMAN_ISOFORMS.mrna_seq_4',
                        protein='HUMAN_ISOFORMS.protein_seq_4'),
             'MOUSE_SPLICE':
             TableGroup(db='MOUSE_SPLICE', suffix='jan02',
                        clusters='cluster_jan02',
                        exons='MOUSE_ISOFORMS.exon_form_2',
                        splices='splice_verification_jan02',
                        genomic='genomic_cluster_jan02',
                        mrna='MOUSE_ISOFORMS.mrna_seq_2',
                        protein='MOUSE_ISOFORMS.protein_seq_2'),
             'MOUSE_SPLICE_03':
             TableGroup(db='MOUSE_SPLICE_03', suffix='JUN03',
                        clusters='cluster_JUN03', exons='exon_formJUN03',
                        splices='splice_verification_JUN03',
                        genomic='genomic_cluster_JUN03', mrna='mrna_seqJUN03',
                        protein='protein_seqJUN03')}


def getUserCursor(db):
    'get a cursor as the current user'
    db = MySQLdb.connect(db=db, read_default_file=os.environ['HOME']
                         + '/.my.cnf', compress=True)
    return db.cursor()


def getSpliceGraphFromDB(dbgroup, loadAll=False):
    """load data from MySQL using the designated database table group.
    If loadAll true, then load the entire splice graph into memory."""
    cursor = getUserCursor(dbgroup.db)
    import sys
    print >>sys.stderr, 'Reading database schema...'
    idDict = {}
    tables = describeDBTables(dbgroup.db, cursor, idDict)
    if hasattr(dbgroup, 'suffix'):
        # Get a set of tables ending in specified suffix
        # and create an index of their primary keys
        tables = suffixSubset(tables, dbgroup.suffix)
        idDict = indexIDs(tables)
    for t in dbgroup.values():
        # This table comes from another database...
        if t is not None and '.' in t and t not in tables:
            tables[t]=SQLTable(t, cursor) # ...so get it from there

    # LOAD DATA & BUILD THE SPLICE GRAPH
    return loadSpliceGraph(tables, dbgroup.clusters, dbgroup.exons,
                           dbgroup.splices, dbgroup.genomic, dbgroup.mrna,
                           dbgroup.protein, loadAll)


def localCopy(localFile, cpCommand):
    'if not already present on local file location, run cpCommand'
    if not os.access(localFile, os.R_OK):
        cmd=cpCommand % localFile
        print 'copying data:', cmd
        exit_code=os.system(cmd)
        if exit_code!=0:
            raise OSError((exit_code, 'command failed: %s' % cmd))
    return localFile




import warnings
warnings.warn('pygr.Data is deprecated.  Use "from pygr import worldbase" \
              instead!', DeprecationWarning, stacklevel=2)

from pygr import worldbase
from metabase import ResourceServer, dumps, OneToManyRelation,\
        OneToOneRelation, ManyToManyRelation, WorldbaseNotPortableError,\
        WorldbaseNotFoundError, WorldbaseMismatchError, WorldbaseEmptyError,\
        WorldbaseReadOnlyError, WorldbaseSchemaError, WorldbaseNoModuleError,\
        ResourceZone

schema = worldbase.schema # ROOT OF OUR SCHEMA NAMESPACE

# PROVIDE TOP-LEVEL NAMES IN OUR RESOURCE HIERARCHY
Bio = worldbase.Bio

getResource = worldbase._mdb # our metabase interface
addResource = worldbase._mdb.add_resource


def addResourceDict(d, layer=None):
    'queue a dict of name:object pairs for saving to specified db layer'
    if layer is not None: # use the named metabase specified by layer
        mdb = worldbase._mdb.zoneDict[layer] # KeyError if layer not found!
    else: # use default MetabaseList
        mdb = worldbase._mdb
    for k, v in d.items(): # queue each resource in the dictionary
        mdb.add_resource(k, v)

addSchema = worldbase._mdb.add_schema
deleteResource = worldbase._mdb.delete_resource
dir = worldbase._mdb.dir


def newServer(*args, **kwargs):
    return ResourceServer(worldbase._mdb, *args, **kwargs)

save = worldbase._mdb.commit
rollback = worldbase._mdb.rollback
list_pending = worldbase._mdb.list_pending
loads = worldbase._mdb.loads
update = worldbase._mdb.update
clear_cache = worldbase._mdb.clear_cache

# TOP-LEVEL NAMES FOR STANDARDIZED LAYERS
here = ResourceZone(getResource, 'here')
my = ResourceZone(getResource, 'my')
system = ResourceZone(getResource, 'system')
subdir = ResourceZone(getResource, 'subdir')
remote = ResourceZone(getResource, 'remote')
MySQL = ResourceZone(getResource, 'MySQL')

__all__ = ('Bio', 'schema', 'getResource', 'addResource', 'addSchema',
           'deleteResource', 'dir', 'newServer', 'save', 'rollback',
           'list_pending', 'loads', 'dumps', 'update', 'clear_cache',
           'OneToManyRelation', 'ManyToManyRelation',
           'OneToOneRelation', 'WorldbaseNotPortableError',
           'WorldbaseNotFoundError', 'WorldbaseMismatchError',
           'WorldbaseEmptyError', 'WorldbaseReadOnlyError',
           'WorldbaseSchemaError', 'WorldbaseNoModuleError',
           'here', 'my', 'system', 'subdir', 'remote', 'MySQL')




import anydbm
import shelve
import sys
import UserDict
import logger


class WrongFormatError(IOError):
    'attempted to open db with the wrong format e.g. btree vs. hash'
    pass


class NoSuchFileError(IOError):
    'file does not exist!'
    pass


class PermissionsError(IOError):
    'inadequate permissions for requested file'
    pass


class ReadOnlyError(PermissionsError):
    'attempted to open a file for writing, but no write permission'
    pass


def open_anydbm(*args, **kwargs):
    'trap anydbm.error message and transform to our consistent exception types'
    try:
        return anydbm.open(*args, **kwargs)
    except ImportError, e:
        if str(e).endswith('bsddb') and bsddb:
            # This almost certainly means dbhash tried to import bsddb
            # on a system with only bsddb3 working correctly. In that case,
            # simply do ourselves what anydbm would have done.
            # FIXME: explicitly check if dbhash raised this exception?
            return bsddb.hashopen(*args, **kwargs)
        raise
    except anydbm.error, e:
        msg = str(e)
        if msg.endswith('new db'):
            raise NoSuchFileError(msg)
        elif msg.startswith('db type'):
            raise WrongFormatError(msg)
        raise


try: # detect whether bsddb module available and working...
    import bsddb
    try:
        bsddb.db
    except AttributeError:
        raise ImportError
except ImportError:
    try:    # maybe the external bsddb3 will work instead...
        import bsddb3
        try:
            bsddb3.db
        except AttributeError:
            raise ImportError
        bsddb = bsddb3
    except ImportError: # ...nope
        bsddb = None


def open_bsddb(filename, flag='r', useHash=False, mode=0666):
    """open bsddb index instead of hash by default.
    useHash=True forces it to use anydbm default (i.e. hash) instead.
    Also gives more meaningful error messages."""
    try: # 1ST OPEN AS BTREE
        if useHash: # FORCE IT TO USE HASH INSTEAD OF BTREE
            return open_anydbm(filename, flag)
        else:
            return bsddb.btopen(filename, flag, mode)
    except bsddb.db.DBAccessError: # HMM, BLOCKED BY PERMISSIONS
        if flag=='c' or flag=='w': # TRY OPENING READ-ONLY
            try:
                ifile = file(filename)
            except IOError:
                # Hmm, not even readable. Raise a generic permission error.
                raise PermissionsError('insufficient permissions \
to open file: ' + filename)
            ifile.close()
            # We can read the file, so raise a ReadOnlyError.
            raise ReadOnlyError('file is read-only: '+ filename)
        else: # r OR n FLAG: JUST RAISE EXCEPTION
            raise PermissionsError('insufficient permissions to open file: '
                                   + filename)
    except bsddb.db.DBNoSuchFileError:
        raise NoSuchFileError('no file named: ' + filename)
    except bsddb.db.DBInvalidArgError: # NOT A BTREE FILE...
        try:
            if useHash: # NO POINT IN TRYING HASH YET AGAIN...
                raise bsddb.db.DBInvalidArgError
            # fallback to using default: hash file
            return open_anydbm(filename, flag)
        except bsddb.db.DBInvalidArgError:
            raise WrongFormatError('file does not match expected \
shelve format: ' + filename)


def open_index(filename, flag='r', useHash=False, mode=0666):
    if bsddb is None:
        d = open_anydbm(filename, flag)
        if not useHash:
            logger.warn('Falling back to hash index: unable to import bsddb')
        return d
    return open_bsddb(filename, flag, useHash, mode)


def iter_gdbm(db):
    'iterator for gdbm objects'
    k = db.firstkey()
    while k is not None:
        yield k
        k = db.nextkey(k)


class _ClosedDict(UserDict.DictMixin):
    """This dummy class exists solely to raise a clear error msg if accessed.
    Copied from the Python 2.6 shelve.py """

    def closed(self, *args):
        raise ValueError('invalid operation on closed shelf')
    __getitem__ = __setitem__ = __delitem__ = keys = closed

    def __repr__(self):
        return '<Closed Dictionary>'


class BetterShelf(shelve.Shelf):
    """Shelf subclass that fixes its horrible iter implementation.
    """

    def __iter__(self):
        'avoid using iter provided by shelve/DictMixin, which loads all keys!'
        try:
            return iter(self.dict)
        except TypeError: # gdbm lacks __iter__ method, so try iter_gdbm()
            exc_type, exc_value, exc_traceback = sys.exc_info()
            try:
                self.dict.firstkey
            except AttributeError: # evidently not a gdbm dict
                raise TypeError('''cannot iterate over this dictionary.
This means that you do not have bsddb, bsddb3, or gdbm available for use by
the 'shelve' module in this Python install.  Please fix this!

Original error message was: %s''' % str(exc_value))
            else: # iterate using gdbm-specific method
                return iter_gdbm(self.dict)

    if sys.version_info < (2, 6): # Python finally added good err msg in 2.6

        def close(self):
            if isinstance(self.dict, _ClosedDict):
                return # if already closed, nothing further to do...
            shelve.Shelf.close(self) # close Shelf as usual
            self.dict = _ClosedDict() # raises sensible error msg if accessed


def shelve_open(filename, flag='c', protocol=None, writeback=False,
                useHash=False, mode=0666, *args, **kwargs):
    """improved implementation of shelve.open() that won't generate
bogus __del__ warning messages like Python's version does."""
    d = open_index(filename, flag, useHash, mode) # construct Shelf only if OK
    return BetterShelf(d, protocol, writeback, *args, **kwargs)



"""
seqdb contains a set of classes for interacting with sequence databases.

Primary sequence database classes:

  - SequenceDB         - base class for sequence databases
  - SequenceFileDB     - file-based sequence database
  - PrefixUnionDict    - container to combine multiple sequence databases
  - XMLRPCSequenceDB   - XML-RPC-accessible sequence database

Extensions:

  - SeqPrefixUnionDict - extends PrefixUnionDict to automatically add seqs
  - BlastDB            - implements NCBI-style name munging for lookups

Associated sequence classes:

  - FileDBSequence     - sequence associated with a SequenceFileDB
  - XMLRPCSequence     - sequence associated with a XMLRPCSequenceDB

----

SequenceDB provides some basic behaviors for sequence databases:
dictionary behavior, an invert operator interface, and caching for
both sequence objects and sequence intervals.  It also relies on a
'seqInfoDict' attribute to contain summary information about
sequences, so that e.g. slice operations can be done without loading
the entire sequence into memory.  (See below for more info on
seqInfoDict.)

SequenceFileDB extends SequenceDB to contain a file-based database of
sequences.  It constructs a seqLenDict that allows direct on-disk lookup
of sequence intervals.  (See below for more info on seqLenDict.)

PrefixUnionDict provides a unified SequenceDB-like interface to a
collection of sequence databases by combining the database name with
the sequence ID into a new sequence id.  For example, the ID
'genome.chrI' would return the sequence 'chrI' in the 'genome'
database.  This is particularly handy for situations where you want to
have seqdbs of multiple sequence types (DNA, protein, annotations,
etc.) all associated together.

@CTB document XMLRPCSequenceDB.
@CTB document SeqPrefixUnionDict.
@CTB document BlastDB.

----

The seqInfoDict interface
-------------------------

The seqInfoDict attribute of a SequenceDB is a dictionary-like object,
keyed by sequence IDs, with associated values being an information
object containing various attributes.  seqInfoDict is essentially an
optimization that permits other pygr-aware components to access
information *about* sequences without actually loading the entire
sequence.

The only required attribute at the moment is 'length', which is
required by some of the NLMSA code.  However, seqInfoDict is a good
mechanism for the storage of any summary information on a sequence,
and so it may be expanded in the future.

The seqLenDict interface
------------------------

The seqLenDict attribute is specific to a SequenceFileDB, where it
provides a file-backed storage of length and offset sequence metadata.
It is used to implement a key optimization in SequenceFileDB, in which
a sequence's offset within a file is used to read only the required
part of the sequence into memory.  This optimization is particularly
important for large sequences, e.g. chromosomes, where reading the
entire sequence into memory shouldn't be done unless it's necessary.

The seqLenDict is keyed by sequence ID and the associated values are a
2-tuple (length, offset), where the offset indicates the byte offset
within the '.pureseq' index file created for each SequenceFileDB.

get_bound_subclass and the 'self.db' attribute
----------------------------------------------

The SequenceDB constructor calls classutil.get_bound_subclass on its
itemClass.  What does that do, and what is it for?

get_bound_subclass takes an existing class, makes a new subclass of
it, binds the variable 'db' to it, and then calls the _init_subclass
classmethod (if it exists) on the new class.  This has the effect of
creating a new class for each SequenceDB instance, tied specifically
to that instance and initialized by the _init_subclass method.

The main effect of this for SequenceDBs is that for any SequenceDB
descendant, the '.db' attribute is automatically set for each Sequence
retrieved from the database.

CTB: I think this is an optimization?

Caching
-------

@CTB discuss caching.

Pickling sequence databases and sequences
-----------------------------------------

@CTB document pickling issues.
programmer notes:

extending SequenceDB
 - seqInfoDict, itemclass

extending SequenceFileDB
 - seqLenDict
 - using your own itemclass
 - using your own reader

doctests & examples
-------------------

update docs for these classes!

intro:
 - loading a FASTA file
 - using a PUD
   + combining dbs, etc.
   + inverse

Code review issues, short term:

 - @CTB get_bound_subclass stuff refers directly to itemClass to set 'db'.
 - @CTB fix 'import *'s
 - @CTB run lint/checker?
 - @CTB test _create_seqLenDict
 - @CTB XMLRPCSequenceDB vs SequenceFileDB

Some long term issues:

 - it should be possible to remove _SeqLenDictSaver and just combine its
   functionality with _store_seqlen_dict.  --titus 3/21/09

"""

from __future__ import generators
import sys
import os
import UserDict
import weakref

from sequence import *                  # @CTB
from sqlgraph import *                  # @CTB
import classutil
from annotation import AnnotationDB, AnnotationSeq, AnnotationSlice, \
     AnnotationServer, AnnotationClient
import logger
import seqfmt

from dbfile import NoSuchFileError


####
#
# SequenceDB and associated support classes.
#

class _SequenceDBInverse(object):
    """Implements __inverse__ on SequenceDB objects, returning seq name."""

    def __init__(self, db):
        self.db = db

    def __getitem__(self, seq):
        return seq.pathForward.id

    def __contains__(self, seq):
        try:
            return seq.pathForward.db is self.db
        except AttributeError:
            return False


class SequenceDB(object, UserDict.DictMixin):
    """Base class for sequence databases.

    SequenceDB provides a few basic (base) behaviors:
      - dict-like interface to sequence objects each with an ID
      - the ~ (invert) operator returns an 'inverted' database, which is a
        dict-like object that returns sequence names when given a sequence.
      - weakref-based automatic flushing of seq objects no longer in use;
        use autoGC=0 to turn this off.
      - cacheHint() system for caching a given set of sequence
        intervals associated with an owner object, which are flushed
        from cache if the owner object is garbage-collected.

    For subclassing, note that self.seqInfoDict must be set before
    SequenceDB.__init__ is called!

    """
    # class to use for database-linked sequences; no default.
    itemClass = None
    # class to use for sequence slices; see sequence.SeqPath.classySlice.
    itemSliceClass = SeqDBSlice

    # pickling methods & what attributes to pickle.
    __getstate__ = classutil.standard_getstate
    __setstate__ = classutil.standard_setstate
    _pickleAttrs = dict(autoGC=0)

    # define ~ (invert) operator to return a lazily-created _SequenceDBInverse
    __invert__ = classutil.lazy_create_invert(_SequenceDBInverse)

    def __init__(self, autoGC=True, dbname=None, **kwargs):
        """Initialize seq db from filepath or ifile."""
        if autoGC: # automatically garbage collect unused objects
            self._weakValueDict = classutil.RecentValueDictionary(autoGC)
        else:
            self._weakValueDict = {}    # object cache @CTB not tested
        self.autoGC = autoGC

        # override itemClass and itemSliceClass if specified
        self.itemClass = kwargs.get('itemClass', self.itemClass)
        self.itemSliceClass = kwargs.get('itemSliceClass', self.itemSliceClass)

        if self.itemClass is None:
            raise TypeError("must supply itemClass to SequenceDB")

        # get a copy we can modify w/o side effects and bind itemClass.
        kwargs = kwargs.copy()
        kwargs['db'] = self
        classutil.get_bound_subclass(self, 'itemClass', dbname,
                                     subclassArgs=kwargs)

        # guess the sequence type
        self._set_seqtype()

    def __hash__(self):
        """Define a hash function to allow this object to be used as a key."""
        return id(self)

    def _set_seqtype(self):
        """Guess the seqtype from 100 chars of 1st seq if not already known."""
        seqtype = getattr(self, '_seqtype', None)
        if seqtype is not None:
            return

        for seqID in self: # get an iterator
            seq = self[seqID] # get the 1st sequence
            ch100 = str(seq[:100])
            self._seqtype = guess_seqtype(ch100)
            break # only process the 1st sequence!!!

    _cache_max=10000                    # @CTB move? make settable?

    def cacheHint(self, ivalDict, owner):
        """Save a cache hint dict: {id: (start, stop)}.

        @CTB document!
        """
        d={}
        # @CTB refactor, test
        # Build the cache dictionary for owner.
        for id, ival in ivalDict.items():
            if ival[0] < 0: # FORCE IVAL INTO POSITIVE ORIENTATION
                ival=(-ival[1], -ival[0])        # @CTB untested
            if ival[1]-ival[0] > self._cache_max: # TRUNCATE EXCESSIVE LENGTH
                ival=(ival[0], ival[0] + self._cache_max) # @CTB untested
            d[id]=[ival[0], ival[1]]
        try:
            self._cache[owner] = d # ADD TO EXISTING CACHE
        except AttributeError:
            self._cache = weakref.WeakKeyDictionary()  # AUTOMATICALLY REMOVE
            self._cache[owner] = d # FROM CACHE IF owner GOES OUT OF SCOPE

    def strsliceCache(self, seq, start, stop):
        """Get strslice using cache hints, if any available."""
        try:
            cacheDict=self._cache
        except AttributeError:
            raise IndexError('no cache present')
        for owner, d in cacheDict.items():
            try:
                ival = d[seq.id]
            except KeyError:
                continue # NOT IN THIS CACHE, SO SKIP  @CTB untested
            ival_start, ival_stop = ival[:2]
            if start >= ival_start and stop <= ival_stop: # CONTAINED IN ival
                try:
                    s = ival[2] # get seq string from our cache
                except IndexError: # use strslice() to retrieve from storage
                    s = seq.strslice(ival_start, ival_stop, useCache=False)
                    ival.append(s)
                    try: # does owner want to reference this cached seq?
                        save_f = owner.cache_reference
                    except AttributeError:
                        pass # no, so nothing to do
                    else: # let owner control caching in our _weakValueDict
                        save_f(seq)     # # @CTB untested
                return s[start - ival_start:stop - ival_start]
        raise IndexError('interval not found in cache') # @CTB untested

    # these methods should all be implemented on all SequenceDBs.
    def close(self):
        pass # subclass should implement closing of its open resources!

    def __iter__(self):
        return iter(self.seqInfoDict)

    def iteritems(self):
        for seqID in self:
            yield seqID, self[seqID]

    def __len__(self):
        return len(self.seqInfoDict)

    def __getitem__(self, seqID):
        """Retrieve sequence by id, using cache if available."""
        try: # for speed, default case (cache hit) should return immediately
            return self._weakValueDict[seqID]
        except KeyError: # not in cache?  try loading.
            try:
                s = self.itemClass(self, seqID)
            except KeyError:
                raise KeyError("no key '%s' in database %s"
                               % (seqID, repr(self)))
            self._weakValueDict[seqID] = s # save in cache.
            return s

    def keys(self):
        return self.seqInfoDict.keys()

    def __contains__(self, key):
        return key in self.seqInfoDict

    def __repr__(self):
        return "<%s '%s'>" % (self.__class__.__name__,
                              self.itemClass.__class__.__name__)

    def clear_cache(self):
        """Empty the cache."""
        self._weakValueDict.clear()

    # these methods should not be implemented for read-only database.
    clear = setdefault = pop = popitem = copy = update = \
            classutil.read_only_error


####
#
# FileDBSequence and SequenceFileDB, and associated support classes.
#

class _FileDBSeqDescriptor(object):
    """Descriptor to retrieve entire sequence from seqLenDict."""

    def __get__(self, obj, objtype):
        length = obj.db.seqLenDict[obj.id][0]
        return obj.strslice(0, length)


class FileDBSequence(SequenceBase):
    """Default sequence object for file-based storage mechanism.

    See SequenceFileDB for the associated database class.

    In general, you should not create objects from this class directly;
    retrieve them from SequenceFileDB objects, instead.

    NOTE: 'self.db' is attached to all instances of this class that come
    from a particular database by 'classutil.get_bound_subclass'.

    """
    seq = _FileDBSeqDescriptor()        # dynamically retrieve 'seq'.
    __reduce__ = classutil.item_reducer # for pickling purposes.

    def __init__(self, db, id):
        self.id = id
        SequenceBase.__init__(self)
        if self.id not in self.db.seqLenDict:
            raise KeyError('sequence %s not in db %s' % (self.id, self.db))

    def __len__(self):
        """Unpack this sequence's length from the seqLenDict."""
        return self.db.seqLenDict[self.id][0]

    def strslice(self, start, end, useCache=True):
        """Access slice of a sequence efficiently, using seqLenDict info."""
        if useCache:                    # If it's in the cache, use that!
            try:
                return self.db.strsliceCache(self, start, end)
            except IndexError:
                pass

        return self.db.strslice(self.id, start, end)


class SequenceFileDB(SequenceDB):
    """Main class for file-based storage of a sequence database.

    By default, SequenceFileDB uses a seqLenDict, a.k.a. a shelve
    index of sequence lengths and offsets, to retrieve sequence slices
    with fseek.  Thus entire chromosomes (for example) do not have to
    be loaded to retrieve a subslice.

    Takes one required argument, 'filepath', which should be the name
    of a FASTA file (or a file whose format is understood by your
    custom reader; see 'reader' kw arg, and the _store_seqlen_dict
    function).

    The SequenceFileDB seqInfoDict interface is a wrapper around the
    seqLenDict created by the __init__ function.

    """
    itemClass = FileDBSequence

    # copy _pickleAttrs and add 'filepath'
    _pickleAttrs = SequenceDB._pickleAttrs.copy()
    _pickleAttrs['filepath'] = 0

    def __init__(self, filepath, reader=None, **kwargs):
        # make filepath a pickleable attribute.
        self.filepath = classutil.SourceFileName(str(filepath))

        fullpath = self.filepath + '.seqlen'
        # build the seqLenDict if it doesn't already exist
        try:
            seqLenDict = classutil.open_shelve(fullpath, 'r')
        except NoSuchFileError:
            seqLenDict = self._create_seqLenDict(fullpath, filepath, reader)

        self.seqLenDict = seqLenDict
        self.seqInfoDict = _SeqLenDictWrapper(self) # standard interface

        # initialize base class.
        dbname = os.path.basename(filepath)
        SequenceDB.__init__(self, filepath=filepath, dbname=dbname, **kwargs)

    def close(self):
        '''close our open shelve index file and _pureseq...'''
        self.seqLenDict.close()
        try:
            do_close = self._pureseq.close
        except AttributeError:
            pass # _pureseq not open yet, so nothing to do
        else:
            do_close()

    def __repr__(self):
        return "<%s '%s'>" % (self.__class__.__name__, self.filepath)

    def _create_seqLenDict(self, dictpath, seqpath, reader=None):
        """Create a seqLenDict from 'seqpath' and store in 'dictpath'."""
        seqLenDict = classutil.open_shelve(dictpath, 'n')
        try:
            logger.debug('Building sequence length index...')
            _store_seqlen_dict(seqLenDict, seqpath, reader)
        finally:
            seqLenDict.close() # close after writing, no matter what!
        return classutil.open_shelve(dictpath, 'r') # re-open read-only

    def strslice(self, seqID, start, end, useCache=True):
        """Access slice of a sequence efficiently, using seqLenDict info."""
        # Retrieve sequence from the .pureseq file based on seqLenDict
        # information.
        try:
            ifile=self._pureseq
        except AttributeError:
            fullpath = self.filepath + '.pureseq'
            ifile = file(fullpath, 'rb')
            self._pureseq = ifile

        # Now, read in the actual slice.
        offset = self.seqLenDict[seqID][1]
        ifile.seek(offset + start)
        return ifile.read(end - start)


# Some support classes for the SeqLenDict mechanism.

class BasicSeqInfo(object):
    """Wrapper to provide the correct seqInfoDict-style object information.

    This boils down to providing id, db, length, and possibly offset.

    """

    def __init__(self, seqID, seqDB, length=None):
        self.id = seqID
        self.db = seqDB
        if length is None:
            self.length = len(seqDB[seqID]) # generic but possibly slow
        else:
            self.length = length


class _SeqLenObject(BasicSeqInfo):
    """Wrapper for use with a seqLenDict """

    def __init__(self, seqID, seqDB):
        length, self.offset = seqDB.seqLenDict[seqID]
        BasicSeqInfo.__init__(self, seqID, seqDB, length)


class BasicSeqInfoDict(object, UserDict.DictMixin):
    """Wrapper around SequenceDB.seqLenDict to provide seqInfoDict behavior.
    This basic version just gets the length from the sequence object itself.
    """
    itemClass = BasicSeqInfo

    def __init__(self, db):
        self.seqDB = db

    def __getitem__(self, k):
        return self.itemClass(k, self.seqDB)

    def __len__(self):
        return len(self.seqDB.seqLenDict)

    def __iter__(self):
        return iter(self.seqDB.seqLenDict)

    def keys(self):
        return self.seqDB.seqLenDict.keys()


class _SeqLenDictWrapper(BasicSeqInfoDict):
    """
    The default storage mechanism for sequences implemented by FileDBSequence
    and SequenceFileDB puts everything in seqLenDict, a shelve index of
    lengths and offsets.  This class wraps that dictionary to provide the
    interface that SequenceDB expects to see.

 """
    itemClass = _SeqLenObject


class _SeqLenDictSaver(object):
    """Support for generic reading functions, called by _store_seqlen_dict.

    This allows you to specify your own 'reader' function when
    constructing a FileSequenceDB, e.g. so that you could read
    something other than FASTA files into a seqLenDict.  Pass in this
    function as the 'reader' kwarg to FileSequenceDB.

    Custom reader functions should take a file handle and a filename,
    and return a list of sequence info objects with 'id', 'length',
    and 'sequence' attributes for each sequence in the given
    file/filename.  _SeqLenDictSaver will then construct a '.pureseq'
    file containing the concatenated sequences and fill in the
    seqLenDict appropriately.

    """

    def __init__(self, reader):
        self.reader = reader

    def __call__(self, d, ifile, filename):
        offset = 0L
        pureseq_fp = file(filename + '.pureseq', 'wb')
        try:
            for o in self.reader(ifile, filename):
                # store the length & offset in the seqLenDict
                d[o.id] = o.length, offset
                offset += o.length
                if o.length != len(o.sequence):
                    raise ValueError('length does not match sequence: %s,%d'
                                     % (o.id, o.length))
                pureseq_fp.write(o.sequence)
        finally:
            pureseq_fp.close()


def _store_seqlen_dict(d, filename, reader=None, mode='rU'):
    """Store sequence lengths in a dictionary, e.g. a seqLenDict.

    Used by SequenceFileDB._create_seqLenDict.

    The 'reader' function implements a custom sequence format reader;
    by default, _store_seqlen_dict uses seqfmt.read_fasta_lengths,
    which reads FASTA-format files.  See _SeqLenDictSaver for
    information on building a custom 'reader', and see the seqdb docs
    for an example.

    """
    # if a custom reader function was passed in, run that.
    builder = seqfmt.read_fasta_lengths
    if reader is not None:
        builder = _SeqLenDictSaver(reader)

    ifile = file(filename, mode)
    try:
        builder(d, ifile, filename) # run the builder on our sequence set
    finally:
        ifile.close()


####
#
# class PrefixUnionDict and associated support classes.
#

class _PrefixUnionDictInverse(object):
    """Provide inverse (~) operator behavior for PrefixUnionDicts.

    This enables ~pud to return a database that, given a sequence
    object, returns the corresponding key (prefix.id) to retrieve that
    sequence object in the pud.

    """

    def __init__(self, db):
        self.db = db

    def __getitem__(self, ival):
        seq = ival.pathForward # get the top-level sequence object
        try: # for speed, normal case should execute immediately
            prefix = self.db.dicts[seq.db]
        except KeyError:
            try:
                # @CTB abstraction boundary violation! keep? how test?
                if seq.pathForward._anno_seq.db in self.db.dicts:
                    raise KeyError('''\
this annotation is not in the PrefixUnion, but its sequence is.
You can get that using its \'sequence\' attribute.''')
            except AttributeError:
                pass
            raise KeyError('seq.db not in PrefixUnionDict')

        return prefix + self.db.separator + str(seq.id)

    def __contains__(self, seq):
        try:
            return seq.pathForward.db in self.db.dicts
        except AttributeError:
            return False


class _PrefixUnionMemberDict(object, UserDict.DictMixin):
    """
    @CTB confusing/inappropriate use of a dict interface! keep??
    @CTB document.
    'd[prefix]=value; d[k] returns value if k is a member of prefix'
    """

    def __init__(self, puDict, default=None,
                 attrMethod=lambda x: x.pathForward.db):
        self.values = {}
        self.puDict = puDict
        self._attrMethod = attrMethod
        if default is not None:
            self.default = default      # @CTB can we use setdefault for this?

    def keys(self):
        return self.puDict.prefixDict.keys()

    possibleKeys = keys                 # legacy interface (?)

    def __setitem__(self, k, v):
        if k not in self.puDict.prefixDict:
            raise KeyError('key %s is not a valid union prefix string!' % k)
        new_k = self.puDict.prefixDict[k]
        self.values[new_k] = v

    def __getitem__(self, k):
        try:
            db = self._attrMethod(k)
        except AttributeError:
            raise TypeError('wrong key type? _attrMethod() failed.')

        if db not in self.values:
            try:
                return self.default
            except AttributeError:      # no default value - raise KeyError.
                raise KeyError('key not a member of this union!')

        return self.values[db]


class PrefixUnionDict(object, UserDict.DictMixin):
    """Interface to a set of sequence DBs, each assigned a unique prefix.

    For example, the sequence ID 'foo.bar' would unpack to ID 'bar' in
    the dictionary associated with the prefix 'foo'.  This is a useful
    way to combine disparate seqdbs into a single db, without actually
    altering the individual seqdbs.

    PrefixUnionDicts can be created in one of two ways: either
      - pass in a dictionary containing prefix-to-seqdb mappings as
        'prefixDict', or
      - pass in a header file containing the information necessary to create
        such a dictionary.

    In the latter case, see the 'writeHeaderFile' method for format
    information.  The optional kwarg 'trypath' contains a list of
    directories to search for the database file named in each line.
    The optional kwarg 'dbClass' specifies the database class to use
    to load each sequence file; it defaults to SequenceFileDB.

    The default ID separator is '.'; use the 'separator' kwarg to
    change it.

    @CTB trypath => trypaths?

    """
    # define ~ (invert) operator to return a lazily-created _PUDInverse.
    __invert__ = classutil.lazy_create_invert(_PrefixUnionDictInverse)

    def __init__(self, prefixDict=None, separator='.', filename=None,
                 dbClass=SequenceFileDB, trypath=None):
        # read union header file
        if filename is not None:
            if prefixDict:
                raise TypeError('''
cannot create with prefixDict and filename both!''')

            if trypath is None:
                trypath = [os.path.dirname(filename)]
            ifile = file(filename, 'rU')
            try:
                it = iter(ifile)
                # Remove leading/trailing CR+LF.
                separator = it.next().strip('\r\n')
                prefixDict = {}
                for line in it:
                    prefix, filepath=line.strip().split('\t')[:2]
                    try:
                        dbfile = classutil.search_dirs_for_file(filepath,
                                                                trypath)
                        db = dbClass(dbfile)
                    except IOError:
                        for db in prefixDict.values():
                            db.close() # close databases before exiting
                        raise IOError('''\
    unable to open database %s: check path or privileges.
    Set 'trypath' to give a list of directories to search.''' % filepath)
                    else:
                        prefixDict[prefix] = db
            finally:
                ifile.close()

        self.separator = separator
        if prefixDict is not None:
            self.prefixDict = prefixDict
        else:
            self.prefixDict = {}

        # also create a reverse mapping
        d = {}
        for k, v in self.prefixDict.items():
            d[v] = k

        self.dicts = d
        self.seqInfoDict = _PUDSeqInfoDict(self) # supply standard interface

    def format_id(self, prefix, seqID):
        return prefix + self.separator + seqID

    def get_prefix_id(self, k):
        """Subdivide a key into a prefix and ID using the given separator."""
        try:
            t = k.split(self.separator, 2) # split into no more than 3 fields
        except AttributeError:
            raise KeyError('key should be a string! ' + repr(k))
        l = len(t)
        if l == 2:
            return t
        elif l<2:
            raise KeyError('invalid id format; no prefix: ' + k)
        else: # id contains separator character?
            prefix = t[0] # assume prefix doesn't contain separator @CTB untested
            seqID = k[len(prefix) + 1:] # skip past prefix
            return prefix, seqID

    def get_subitem(self, d, seqID):
        # try int key first
        try:
            return d[int(seqID)]
        except (ValueError, KeyError, TypeError):
            pass

        # otherwise, use default (str) key
        try:
            return d[seqID]
        except KeyError:
            raise KeyError("no key '%s' in %s" % (seqID, repr(d)))

    def __getitem__(self, k):
        """For 'foo.bar', return 'bar' in dict associated with prefix 'foo'"""
        prefix, seqID = self.get_prefix_id(k)
        try:
            d = self.prefixDict[prefix]
        except KeyError, e:
            raise KeyError("no key '%s' in %s" % (k, repr(self)))
        return self.get_subitem(d, seqID)

    def __contains__(self, k):
        """Is the given ID in our PrefixUnionDict?"""
        # try it out as an ID.
        if isinstance(k, str):
            try:
                (prefix, id) = self.get_prefix_id(k)
                return id in self.prefixDict[prefix]
            except KeyError:
                return False

        # otherwise, try treating key as a sequence.
        # @CTB inconsistent with 'getitem'.
        try:
            db = k.pathForward.db
        except AttributeError:
            raise AttributeError('key must be a sequence with db attribute!')
        return db in self.dicts

    def has_key(self, k):
        return self.__contains__(k)

    def __iter__(self):
        for p, d in self.prefixDict.items():
            for id in d:
                yield self.format_id(p, id)

    def keys(self):
        return list(self.iterkeys())

    def iterkeys(self):
        return iter(self)

    def iteritems(self):
        for p, d in self.prefixDict.items():
            for id, seq in d.iteritems():
                yield self.format_id(p, id), seq

    def getName(self, ival):
        """For a given sequence, return a fully qualified name, 'prefix.id'."""
        seq = ival.pathForward # get the top-level sequence object
        return self.dicts[seq.db] + self.separator + seq.id

    def newMemberDict(self, **kwargs):  # @CTB not used; necessary?
        """return a new member dictionary (empty)"""
        return _PrefixUnionMemberDict(self, **kwargs)

    def writeHeaderFile(self, filename):  # @CTB not used; necessary?
        """Save a header file, suitable for later re-creation."""
        ifile = file(filename, 'w')
        print >>ifile, self.separator
        for k, v in self.prefixDict.items():
            try:
                print >>ifile, '%s\t%s' % (k, v.filepath)
            except AttributeError:
                raise AttributeError('''\
seq db '%s' has no filepath; you may be able to save this to worldbase,
but not to a text HeaderFile!''' % k)
        ifile.close()

    def __len__(self):
        n=0
        for db in self.dicts:
            n += len(db)
        return n

    def cacheHint(self, ivalDict, owner=None):  # @CTB untested
        '''save a cache hint dict of {id:(start,stop)}'''
        d={}
        # extract separate cache hint dict for each prefix
        for longID, ival in ivalDict.items():
            prefix, seqID = self.get_prefix_id(longID)
            d.setdefault(prefix, {})[seqID] = ival
        for prefix, seqDict in d.items():
            try:
                m = self.prefixDict[prefix].cacheHint
            except AttributeError: # subdict can't cacheHint(), so just ignore
                pass
            else:
                # pass cache hint down to subdictionary
                m(seqDict, owner)

    # not clear what this should do for PrefixUnionDict
    copy = setdefault = update = classutil.method_not_implemented

    # these methods should not be implemented for read-only database.
    clear = pop = popitem = classutil.read_only_error


class _PrefixDictInverseAdder(_PrefixUnionDictInverse):
    """Inverse class for SeqPrefixUnionDict; adds sequences when looked up.

    @CTB is getName only used by __getitem__?  Make private?
    """

    def getName(self, seq):
        """Find in or add the given sequence to the inverse of a PUD."""
        try:
            return _PrefixUnionDictInverse.__getitem__(self, seq)
        except AttributeError: # no seq.db?  treat as a user sequence.
            new_id = 'user' + self.db.separator + seq.pathForward.id
            # check to make sure it's already in the user seq db...
            _ = self.db[new_id]
            return new_id

    def __getitem__(self, seq):
        """__getitem__ interface that calls getName."""
        try:
            return self.getName(seq)
        except KeyError:
            if not self.db.addAll:
                raise

            # if we should add, add seq & re-try.
            self.db += seq
            return self.getName(seq)


class SeqPrefixUnionDict(PrefixUnionDict):
    """SeqPrefixUnionDict provides += functionality to add seqs to a PUD.

    See the __iadd__ method for details.

    If addAll is True, then looking a sequence up in the inverse db will
    automatically add it to the PrefixUnionDict.
    """

    __invert__ = classutil.lazy_create_invert(_PrefixDictInverseAdder)

    def __init__(self, addAll=False, **kwargs):
        PrefixUnionDict.__init__(self, **kwargs)

        # override default PrefixUnionDict __invert__ to add sequences;
        # see classutil.lazy_create_invert.
        self.addAll = addAll  # see _PrefixDictInverseAdder behavior.

    def __iadd__(self, k):
        """Add a sequence or database to the PUD, with a unique prefix.

        NOTE: __iadd__ must return self.

        """
        # seq or db already present?
        if k in (~self):
            return self

        db = getattr(k, 'db', None)
        if db is None:                  # annotation sequence?
            db = getattr(k.pathForward, 'db', None) # @CTB untested

        if db is None:  # this is a user sequence, with no container; create.
            if 'user' not in self.prefixDict:
                d = KeepUniqueDict()
                self._add_prefix_dict('user', d)
            else:
                d = self.prefixDict['user']

            # now add the sequence
            d[k.pathForward.id] = k.pathForward
            return self

        # already contain?  nothing to do.
        if db in self.dicts:            # @CTB can this 'if' ever be true?
            return self

        # ok, not present; add, with a unique name.  does it have
        # _persistent_id?
        try:
            name = db._persistent_id.split('.')[-1]
        except AttributeError:          # no; retrieve from filepath?
            name = getattr(db, 'filepath', None)
            if name:                    # got one; clean up.
                name = os.path.basename(name)
                name = name.split('.')[0]
            else:                       # generate one.
                name = 'noname%d' % len(self.dicts)

            if name in self.prefixDict:
                logger.debug('''
It appears that two different sequence databases are being assigned
the same prefix ("%s").  For this reason, the attempted automatic
construction of a PrefixUnionDict for you cannot be completed!  You
should instead construct a PrefixUnionDict that assigns a unique
prefix to each sequence database, and supply it directly as the
seqDict argument to the NLMSA constructor.''' % id)
                raise ValueError('''\
cannot automatically construct PrefixUnionDict''')

        self._add_prefix_dict(name, db)

        return self

    def _add_prefix_dict(self, name, d):
        self.prefixDict[name] = d
        self.dicts[d] = name


class _PUDSeqInfoDict(object, UserDict.DictMixin):
    """A wrapper object supplying a standard seqInfoDict interface for PUDs.

    This class simply provides a standard dict interface that rewrites
    individual sequence IDs into the compound PrefixUnionDict seq IDs
    on the fly.

    """

    def __init__(self, db):
        self.seqDB = db

    def __iter__(self):
        return iter(self.seqDB)

    def keys(self):
        return list(self.iterkeys())

    def iterkeys(self):
        for (k, v) in self.iteritems():
            yield k

    def itervalues(self):
        for (k, v) in self.iteritems():
            yield v

    def iteritems(self):
        for p, d in self.seqDB.prefixDict.items():
            for seqID, info in d.seqInfoDict.iteritems():
                yield self.seqDB.format_id(p, seqID), info

    def __getitem__(self, k):
        prefix, seqID = self.seqDB.get_prefix_id(k)
        db = self.seqDB.prefixDict[prefix]
        return self.seqDB.get_subitem(db.seqInfoDict, seqID)

    def has_key(self, k):
        return k in self.seqDB

#
# @CTB stopped review here. ###################################################
#

class BlastDB(SequenceFileDB):          # @CTB untested?
    '''Deprecated interface provided for backwards compatibility.
    Provides blast() and megablast() methods for searching your seq db.
    Instead of this, you should use the blast.BlastMapping, which provides
    a graph interface to BLAST, or MegablastMapping for megablast.'''

    def __reduce__(self): # provided only for compatibility w/ 0.7 clients
        return (classutil.ClassicUnpickler, (self.__class__,
                                             self.__getstate__()))

    def __init__(self, filepath=None, blastReady=False, blastIndexPath=None,
                 blastIndexDirs=None, **kwargs):
        """format database and build indexes if needed. Provide filepath
        or file object"""
        SequenceFileDB.__init__(self, filepath, **kwargs)

    def __repr__(self):
        return "<BlastDB '%s'>" % (self.filepath)

    def blast(self, seq, al=None, blastpath='blastall',
              blastprog=None, expmax=0.001, maxseq=None, verbose=True,
              opts='', **kwargs):
        'run blast with the specified parameters, return NLMSA alignment'
        blastmap = self.formatdb()
        return blastmap(seq, al, blastpath, blastprog, expmax, maxseq,
                        verbose, opts, **kwargs)

    def megablast(self, seq, al=None, blastpath='megablast', expmax=1e-20,
                  maxseq=None, minIdentity=None, maskOpts='-U T -F m',
                  rmPath='RepeatMasker', rmOpts='-xsmall',
                  verbose=True, opts='', **kwargs):
        'run megablast with the specified parameters, return NLMSA alignment'
        from blast import MegablastMapping
        blastmap = self.formatdb(attr='megablastMap',
                                 mapClass=MegablastMapping)
        return blastmap(seq, al, blastpath, expmax, maxseq, minIdentity,
                        maskOpts, rmPath, rmOpts, verbose, opts, **kwargs)

    def formatdb(self, filepath=None, attr='blastMap', mapClass=None):
        'create a blast mapping object if needed, and ensure it is indexed'
        try: # see if mapping object already exists
            blastmap = getattr(self, attr)
        except AttributeError:
            if mapClass is None: # default: BlastMapping
                from blast import BlastMapping
                mapClass = BlastMapping
            blastmap = mapClass(self)
            setattr(self, attr, blastmap) # re-use this in the future
        blastmap.formatdb(filepath) # create index file if not already present
        return blastmap


class BlastDBXMLRPC(BlastDB):
    'XMLRPC server wrapper around a standard BlastDB'
    xmlrpc_methods = dict(getSeqLen=0, get_strslice=0, getSeqLenDict=0,
                          get_db_size=0, get_seqtype=0,
                          strslice='get_strslice')

    def getSeqLen(self, id):
        'get sequence length, or -1 if not found'
        try:
            return len(self[id])
        except KeyError:
            return -1  # SEQUENCE OBJECT DOES NOT EXIST

    def getSeqLenDict(self):
        'return seqLenDict over XMLRPC'
        d = {}
        for k, v in self.seqLenDict.items():
            d[k] = v[0], str(v[1]) # CONVERT TO STR TO ALLOW OFFSET>2GB
        return d # XML-RPC CANNOT HANDLE INT > 2 GB, SO FORCED TO CONVERT...

    def get_db_size(self):
        return len(self)

    def get_strslice(self, id, start, stop):
        '''return string sequence for specified interval
        in the specified sequence'''
        if start < 0: # HANDLE NEGATIVE ORIENTATION
            return str((-(self[id]))[-stop:-start])
        else: # POSITIVE ORIENTATION
            return str(self[id][start:stop])

    def get_seqtype(self):
        return self._seqtype


class XMLRPCSequence(SequenceBase):
    "Represents a sequence in a blast database, accessed via XMLRPC"

    def __init__(self, db, id):
        self.length = db.server.getSeqLen(id)
        if self.length <= 0:
            raise KeyError('%s not in this database' % id)
        self.id = id
        SequenceBase.__init__(self)

    def strslice(self, start, end, useCache=True):
        "XMLRPC access to slice of a sequence"
        if useCache:
            try:
                return self.db.strsliceCache(self, start, end)
            except IndexError: # NOT FOUND IN CACHE
                pass # JUST USE OUR REGULAR XMLRPC METHOD
        # Get from XMLRPC.
        return self.db.server.get_strslice(self.id, start, end)

    def __len__(self):
        return self.length


class XMLRPCSeqLenDescr(object):
    'descriptor that returns dictionary of remote server seqLenDict'

    def __init__(self, attr):
        self.attr = attr

    def __get__(self, obj, objtype):
        '''only called if attribute does not already exist. Saves result
        as attribute'''
        d = obj.server.getSeqLenDict()
        for k, v in d.items():
            d[k] = v[0], int(v[1]) # CONVERT OFFSET STR BACK TO INT
        obj.__dict__[self.attr] = d # PROVIDE DIRECTLY TO THE __dict__
        return d


class XMLRPCSequenceDB(SequenceDB):
    'XMLRPC client: access sequence database over XMLRPC'
    itemClass = XMLRPCSequence # sequence storage interface
    seqLenDict = XMLRPCSeqLenDescr('seqLenDict') # INTERFACE TO SEQLENDICT

    def __init__(self, url, name, *args, **kwargs):
        import coordinator
        self.server = coordinator.get_connection(url, name)
        self.url = url
        self.name = name
        self.seqInfoDict = _SeqLenDictWrapper(self)
        SequenceDB.__init__(self, *args, **kwargs)

    def __reduce__(self): # provided only for compatibility w/ 0.7 clients
        return (classutil.ClassicUnpickler, (self.__class__,
                                             self.__getstate__()))

    def __getstate__(self): # DO NOT pickle self.itemClass! We provide our own.
        return dict(url=self.url, name=self.name) # just need XMLRPC info

    def __len__(self):
        return self.server.get_db_size()

    def __contains__(self, k):
        if self.server.getSeqLen(k)>0:
            return True
        else:
            return False

    def _set_seqtype(self):
        'efficient way to determine sequence type of this database'
        try: # if already known, no need to do anything
            return self._seqtype
        except AttributeError:
            self._seqtype = self.server.get_seqtype()
            return self._seqtype



from __future__ import generators
import math
from nlmsa_utils import CoordsGroupStart, CoordsGroupEnd

# AUTHORS: zfierstadt, leec


def is_line_start(token, line):
    "check whether line begins with token"
    return token == line[:len(token)]


def get_ori_letterunit(start, end, seq, gapchar='-'):
    """try to determine orientation (1 or -1) based on whether start>end,
    and letterunit (1 or 3) depending on the ratio of end-start difference
    vs the actual non-gap letter count.  Returns tuple (ori,letterunit)"""
    if end > start:
        ori = 1
    else:
        ori = -1
    ngap = 0
    for l in seq:
        if l == gapchar:
            ngap += 1
    seqlen = len(seq) - ngap
    if ori * float(end - start) / seqlen > 2.0:
        letterunit = 3
    else:
        letterunit = 1
    return ori, letterunit


class BlastIval(object):

    def __repr__(self):
        return '<BLAST-IVAL: ' + repr(self.__dict__) + '>'


class BlastHitParser(object):
    """reads alignment info from blastall standard output.
    Method parse_file(fo) reads file object fo, and generates tuples
    suitable for BlastIval."""
    gapchar = '-'

    def __init__(self):
        self.hit_id = 0
        self.nline = 0
        self.reset()

    def reset(self):
        "flush any alignment info, so we can start reading new alignment"
        self.query_seq = ""
        self.subject_seq = ""
        self.hit_id += 1

    def save_query(self, line):
        self.query_id = line.split()[1]

    def save_subject(self, line):
        self.subject_id = line.split()[0][1:]

    def save_score(self, line):
        "save a Score: line"
        self.blast_score = float(line.split()[2])
        s = line.split()[7]
        if s[0] == 'e':
            s = '1' + s
        try:
            self.e_value = -math.log(float(s)) / math.log(10.0)
        except (ValueError, OverflowError):
            self.e_value = 300.

    def save_identity(self, line):
        "save Identities line"
        s = line.split()[3][1:]
        self.identity_percent = int(s[:s.find('%')])

    def save_query_line(self, line):
        "save a Query: line"
        c=line.split()
        self.query_end=int(c[3])
        if not self.query_seq:
            self.query_start=int(c[1])
            # Handle forward orientation.
            if self.query_start < self.query_end:
                self.query_start -= 1
        self.query_seq+=c[2]
        self.seq_start_char=line.find(c[2], 5) # IN CASE BLAST SCREWS UP Sbjct:

    def save_subject_line(self, line):
        "save a Sbjct: line, attempt to handle various BLAST insanities"
        c=line.split()
        if len(c)<4: # OOPS, BLAST FORGOT TO PUT SPACE BEFORE 1ST NUMBER
            # THIS HAPPENS IN TBLASTN... WHEN THE SUBJECT SEQUENCE
            # COVERS RANGE 1-1200, THE FOUR DIGIT NUMBER WILL RUN INTO
            # THE SEQUENCE, WITH NO SPACE!!
            c = ['Sbjct:', line[6:self.seq_start_char]] \
               + line[self.seq_start_char:].split() # FIX BLAST SCREW-UP
        self.subject_end=int(c[3])
        if not self.subject_seq:
            self.subject_start = int(c[1])
            # Handle forward orientation.
            if self.subject_start < self.subject_end:
                self.subject_start -= 1
        self.subject_seq += c[2]
        lendiff = len(self.query_seq) - len(self.subject_seq)
        if lendiff > 0: # HANDLE TBLASTN SCREWINESS: Sbjct SEQ OFTEN TOO SHORT!
            # THIS APPEARS TO BE ASSOCIATED ESPECIALLY WITH STOP CODONS *
            self.subject_seq += lendiff * 'A' # EXTEND TO SAME LENGTH AS QUERY
        elif lendiff < 0 and not hasattr(self, 'ignore_query_truncation'):
            # WHAT THE HECK?!?!  WARN THE USER: BLAST RESULTS ARE SCREWY...
            raise ValueError(
                """BLAST appears to have truncated the Query: sequence
                to be shorter than the Sbjct: sequence:
                Query: %s
                Sbjct: %s
                This should not happen!  To ignore this error, please
                create an attribute ignore_query_truncation on the
                BlastHitParser object.""" % (self.query_seq, self.subject_seq))

    def get_interval_obj(self, q_start, q_end, s_start, s_end,
                         query_ori, query_factor, subject_ori, subject_factor):
        "return interval result as an object with attributes"
        o = BlastIval()
        o.hit_id = self.hit_id
        o.src_id = self.query_id
        o.dest_id = self.subject_id
        o.blast_score = self.blast_score
        o.e_value = self.e_value
        o.percent_id = self.identity_percent
        o.src_ori = query_ori
        o.dest_ori = subject_ori
        query_start = self.query_start+q_start*query_ori*query_factor
        query_end = self.query_start+q_end*query_ori*query_factor
        subject_start = self.subject_start+s_start*subject_ori*subject_factor
        subject_end = self.subject_start+s_end*subject_ori*subject_factor
        if query_start<query_end:
            o.src_start = query_start
            o.src_end = query_end
        else:
            o.src_start = query_end
            o.src_end = query_start
        if subject_start<subject_end:
            o.dest_start = subject_start
            o.dest_end = subject_end
        else:
            o.dest_start = subject_end
            o.dest_end = subject_start
        return o

    def is_valid_hit(self):
        return self.query_seq and self.subject_seq

    def generate_intervals(self):
        "generate interval tuples for the current alignment"
        yield CoordsGroupStart() # bracket with grouping markers

        query_ori, query_factor = get_ori_letterunit(self.query_start,\
                  self.query_end, self.query_seq, self.gapchar)
        subject_ori, subject_factor = get_ori_letterunit(self.subject_start,\
                  self.subject_end, self.subject_seq, self.gapchar)
        q_start= -1
        s_start= -1
        i_query=0
        i_subject=0
        for i in range(len(self.query_seq)): # SCAN ALIGNMENT FOR GAPS
            if self.query_seq[i] == self.gapchar or \
               self.subject_seq[i] == self.gapchar:
                if q_start >= 0: # END OF AN UNGAPPED INTERVAL
                    yield self.get_interval_obj(q_start, i_query,
                                                s_start, i_subject,
                                                query_ori, query_factor,
                                                subject_ori, subject_factor)
                q_start= -1
            elif q_start<0: # START OF AN UNGAPPED INTERVAL
                q_start=i_query
                s_start=i_subject
            if self.query_seq[i]!=self.gapchar: # COUNT QUERY LETTERS
                i_query+=1
            if self.subject_seq[i]!=self.gapchar: # COUNT SUBJECT LETTERS
                i_subject+=1
        if q_start>=0: # REPORT THE LAST INTERVAL
            yield self.get_interval_obj(q_start, i_query,
                                        s_start, i_subject,
                                        query_ori, query_factor,
                                        subject_ori, subject_factor)

        yield CoordsGroupEnd()

    def parse_file(self, myfile):
        "generate interval tuples by parsing BLAST output from myfile"
        for line in myfile:
            self.nline += 1
            if self.is_valid_hit() and \
               (is_line_start('>', line) or is_line_start(' Score =', line) \
                or is_line_start('  Database:', line) \
                or is_line_start('Query=', line)):
                for t in self.generate_intervals(): # REPORT THIS ALIGNMENT
                    yield t # GENERATE ALL ITS INTERVAL MATCHES
                self.reset() # RESET TO START A NEW ALIGNMENT
            if is_line_start('Query=', line):
                self.save_query(line)
            elif is_line_start('>', line):
                self.save_subject(line)
            elif is_line_start(' Score =', line):
                self.save_score(line)
            elif 'Identities =' in line:
                self.save_identity(line)
            elif is_line_start('Query:', line):
                self.save_query_line(line)
            elif is_line_start('Sbjct:', line):
                self.save_subject_line(line)
        if self.nline == 0: # no blast output??
            raise IOError('No BLAST output. Check that blastall is \
                          in your PATH')

if __name__=='__main__':
    import sys
    p=BlastHitParser()
    for t in p.parse_file(sys.stdin):
        print t
